/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    buildPatchResponseCache.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
** DESCRIPTION:
**  Build a cache of patch responses.
**
*****************************************************************************/

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sys/types.h>
#include <sstream>

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#include "win32/dirent.h"
#else
#include <dirent.h>
#endif

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlVision.h"
#include "svlDevel.h"

using namespace std;


// Thread args
struct ProcessFileArgs {
  // idx is the file number within the directory; sample_idx is the number of
  // the output file
  ProcessFileArgs(  const char *outD,
		    const svlDetectorFeatureExtractor * dict,  unsigned idx, const svlPatchUtil & util,
		    unsigned sample_idx) :
      outputDir(outD),  
      dictionary(dict), idx(idx), util(util), sample_idx(sample_idx) {}



  const char *outputDir;

  const svlDetectorFeatureExtractor * dictionary;
  unsigned idx;
  const svlPatchUtil & util;
  unsigned sample_idx;
};

// Thread function
void *processFile(void *voidArgs) {
  //Copy arguments out of the helper struct
  ProcessFileArgs *args = (ProcessFileArgs*) voidArgs;
  const char *outputDir = args->outputDir;
  const svlDetectorFeatureExtractor * dictionary = args->dictionary;
  unsigned idx = args->idx;
  unsigned sample_idx = args->sample_idx;
  const svlPatchUtil & util = args->util;

  vector<IplImage *> images;

  if (!util.getFrame(idx, images))
    return NULL;
            
  // process image
  vector<double> v;
  dictionary->windowResponse(images, v);
            
  // write feature vector
  string outputFilename = string(outputDir) + string("/") + "sample" + toString(sample_idx) + ".txt";

  SVL_LOG(SVL_LOG_VERBOSE, "...writing output to " << outputFilename.c_str());
            
  ofstream ofs(outputFilename.c_str());
  if (ofs.fail()) {
    SVL_LOG(SVL_LOG_WARNING, "...failed");
  } else {
    for (unsigned i = 0; i < v.size(); i++) {
      ofs << v[i] << endl;
    }
  }
  ofs.close();
            
  // Free memory
  for (unsigned i = 0; i < images.size(); i++)
    cvReleaseImage(&(images[i]));

  delete args;

  return NULL;
}

void processBinaryFile(std::ostream& ofs, const svlDetectorFeatureExtractor * dictionary,
                       unsigned idx, const svlPatchUtil& util) {
  vector<IplImage *> images;

  if (!util.getFrame(idx, images))
    return ;

  // process image
  vector<double> v;
  dictionary->windowResponse(images,v);

  // write feature vector
  if (ofs.fail()) {
    SVL_LOG(SVL_LOG_WARNING, "...failed on index " << idx);
  } else {    
    uint32_t sz = v.size();
    ofs.write((const char*)&sz, sizeof(uint32_t));
    for (unsigned i = 0; i < v.size(); i++) {
      ofs.write((const char*)&v[i], sizeof(double));
    }
  }
            
  // Free memory
  for (unsigned i = 0; i < images.size(); i++)
    cvReleaseImage(&(images[i]));
}

void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./buildPatchResponseCache [OPTIONS] <image dir> <output dir> <feature extractor>" << endl;
    cerr << "OPTIONS:" << endl
	 << "  -maxImages <num>   :: maximum training images to cache (default: 10000)" << endl
         << "  -profile           :: turn on SVL code profiling" << endl
         << "  -v                 :: verbose" << endl
	 << "  -dmsize <w> <h>    :: dimensions of the depth map"<< endl
	 << "  -ch <type> <ext>   :: adds a channel" << endl
	 << "  -binary            :: creates binary cache file named <output dir>.bin" << endl
	 << "  -parallel <id> <ct>:: used for distributed jobs on the cluster." << endl
	 << "  -startIndex        :: starts numbering from this index instead of from 0" << endl
	 << "  -augment           :: augments the existing cache dir; overrides -startIndex " << endl
	 << "                        (same as using -startIndex [curr # files in cache dir])" << endl
	 << "  Note: both -augment and -startIndex make sure the TOTAL number of files in the " << endl
	 << "        cache dir doesn't exceed -maxImages, not just the number of files added" << endl
	 << "        in the current run." << endl
	 << "DEPRECATED OPTIONS: "<<endl
	 << "  -ch <n> <c1> ... <c_n> :: n input channels, along with types for each of them"<< endl
	 << "  -imageExt <str>    :: extension of the image data"<< endl
	 << "  -depthMapExt <str> :: extension of the depth map data if any"<< endl
	 << endl;
}


int main(int argc, char *argv[])
{
    // read commandline parameters
    const int NUM_REQUIRED_PARAMETERS = 3;

    //Default values for the channels (transforms of the raw input data)
    vector<svlChannelType> channelTypes;
    vector<string> channelExts;
    
    const char *imageDir = NULL;
    const char *outputDir = NULL;
    const char *extractorFilename = NULL;
    int maxImages = 10000;
    int nChannels = 2;
    //int depthMapWidth = -1;
    //int depthMapHeight = -1;
    string imageExt = string(".jpg");
    string depthMapExt = string(".reconstructed.z.txt");
    bool binary = false;
    int  parallelId = 0;
    int  parallelCount = 0;
    bool augment = false;
    int startIndex = 0;

    svlPatchUtil util(SVL_PATCH_DEFAULT_EDGE);

    char **args = argv + 1;
    while (--argc > NUM_REQUIRED_PARAMETERS) {
      if (util.parseCommandLineItem(argc, args))
	{
	  //do nothing, util handled this flag
	}
      else if (!strcmp(*args, "-binary")) {
        binary = true;
      }
      else if (!strcmp(*args, "-parallel")) {
        parallelId = atoi(*(++args)); argc--;
        parallelCount = atoi(*(++args)); argc--;
        if (parallelId < 0 || parallelId >= parallelCount) {
	  SVL_LOG(SVL_LOG_FATAL, "illegal id number " << parallelId);
        } else if (parallelCount < 0) {
	  SVL_LOG(SVL_LOG_FATAL, "illegal job count " << parallelCount);
        }
      }
      else if(!strcmp(*args, "-augment")) {
	augment = true;
      } else if(!strcmp(*args, "-startIndex")) {
	startIndex = atoi(*(++args)); argc--;
      } else if (!strcmp(*args, "-maxImages")) {
	maxImages = atoi(*(++args)); argc--;
      } else if (!strcmp(*args, "-profile")) {
	svlCodeProfiler::enabled = true;
	svlCodeProfiler::tic(svlCodeProfiler::getHandle("buildPatchResponseCache"));
      }  else if (!strcmp(*args, "-v")) {
	svlLogger::setLogLevel(SVL_LOG_VERBOSE);
      } 
      else {
	SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
      }
      args++;
    }
    
    if (argc != NUM_REQUIRED_PARAMETERS) {
	usage();
	return -1;
    }

    imageDir = args[0];
    outputDir = args[1];
    extractorFilename = args[2];
    
    util.finishCommandLine();

    nChannels = util.numChannels();

    // read dictionary
    const svlDetectorFeatureExtractor *  dictionary = svlDetectorFeatureExtractor::getInstance(extractorFilename);
    
    //stringstream ss;
    //ss << dictionaryFilename << 2;
    //dictionary.write(ss.str().c_str());
	
    SVL_LOG(SVL_LOG_MESSAGE, dictionary->numFeatures() << " features in the feature extractor");

    int width = dictionary->windowWidth();
    int height = dictionary->windowHeight();
    
    svlSoftEdgeMap edgeMapCalculator;
    
    // will be compared to the maxImages parameter
    int numImages = 0;
    if (startIndex > 0 || augment)
      numImages = svlDirectoryListing(outputDir).size();

    if (augment)
      startIndex = numImages;

    util.readDir(imageDir);
    util.setResizeSize(width,height);

    if (!binary) {
      // ------ Output ASCII cache files ------

      // TODO: how many threads?
      svlThreadPool threadPool(8);
      threadPool.start();
      
      // i is the index of the image we're processing, numImages is the number of
      // images in the cache directory currently, i+startIndex is the name of the file
      // we're about to output
      for (unsigned i = 0; i < util.numFrames() && numImages < maxImages; i++, numImages++) {
        //if (parallelCount > 0) { // UNTESTED
        //if ((i % parallelCount) != parallelId) continue ;  // UNTESTED
        //}  // UNTESTED

        threadPool.addJob(processFile,
                          new ProcessFileArgs(outputDir, dictionary, i, util, i + startIndex));
      }
      // TODO: make this contingent on success?
    
      threadPool.finish();
    } else {
      // ------ Output binary cache file ------
      if (augment) {
        SVL_LOG(SVL_LOG_FATAL, "-augment flag and -binary flag are incompatible at present.");
      }

      // erases the possible "/" at the end of the directory name
      string outputDirStr = string(outputDir);
      int pos = outputDirStr.find_last_not_of("/");
      outputDirStr = outputDirStr.substr(0, pos+1);

      std::string jobId = "";
      if (parallelCount > 0) {
        jobId = "." + toString(parallelId);
      }
      std::string fileName = outputDirStr+jobId+".bin";
      std::fstream ofs(fileName.c_str(), std::ios::out|std::ios::binary);
      
      for (unsigned i = 0; i < util.numFrames() && numImages < maxImages; i++, numImages++) {
        if (parallelCount > 0) {
          if ((i % unsigned(parallelCount)) != unsigned(parallelId)) continue ;
        }
        processBinaryFile(ofs, dictionary, i, util);
      }
    }

    svlCodeProfiler::toc(svlCodeProfiler::getHandle("buildPatchResponseCache"));
    svlCodeProfiler::print(cerr);
    
    return 0;
}

