/****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    buildTrainingImageDataset.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
** DESCRIPTION:
**  Constructs a database of positive and negative training images for
**  training an object detector.
**
*****************************************************************************/

#include <cstdlib>
#include <string>
#include <limits>
#include <iomanip>
#include <map>
#include <set>

// OpenCV library
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

// stair vision library
#include "svlBase.h"
#include "svlVision.h"

using namespace std;

#define MAX_NUM_TRIES 100

//Specifies the patch type and file extension of a channel of a sequence
typedef pair<svlPatchDefinitionType, string> ChannelPair;

// randRect -------------------------------------------------------------------
// Given a base size and frame size, this returns a rectangle in the frame that
// is sampled from the distribution consistent with what the classifier will
// see in the sliding windows (essentially, more small, fewer large images).

// TODO: use in main

// TODO: aspect ratio input is redundant, given base size.  Which is better for
// caller?
CvRect randRect(double aspectRatio, CvSize base, CvSize frame, CvRNG *rngState);

//saves an image using a method specific to its patch type
void savePatch(const string & filename, const IplImage * patch, svlPatchDefinitionType type);

//Displays command line options help
void usage();


//Populates "output" with a list of rectangles to turn into negative examples
//Checks that rectangles do not overlap with "objects" (the ground truth labels)
//Draws negative examples from false positives in "objectDetections"
//Uses "imageId" to tell which frome of "objectDetections" to work with
//If "fponly" is true, all negative examples come from false positives
//Generates numNegatives examples per frame
//aspect ratio determined by negAspectRatio
void getNegatives(svlObject2dFrame & output, const svlObject2dFrame & objects, svlObject2dSequence & objectDetections,
		  const string & imageId, const int & fponly, const int & numNegatives, const double & negAspectRatio,
		  CvRNG & rngState, IplImage * image, const int & minWidth, const int & minHeight, const int & negHeight,
		  const float &threshold, const svlObject2dFrame &otherObjects);

//Copies the region r out of image. Does appropriate formatting based on the channel type.
//Saves the clip to a filepath that is a function of the channel extension, base directory, typeName, class prefix, an otionally different type name to use while saving (this is stupid, see todo down by implementation), index of the frame being saved within that class
//Optionally does this a second time with the image flipped and an "f" wedged in the filename (does not increment the frame index)
//Yes, this does have too many arguments. I made it as a spot fix for a section of code that was copy-pasted with minor edits in three places. We could certainly make
//this more graceful though
void handle_clip(IplImage * image, const CvRect & r, svlPatchDefinitionType pType, const string & ext, string baseDirectory, string typeName, const char * prefix, string typeNameForFile, unsigned int idx, bool bIncludeFlipped, int resizeWidth, int resizeHeight);


// Main ----------------------------------------------------------------------

int main(int argc, char *argv[])
{
  vector<ChannelPair> channels;

  CvRNG rngState(0xffffffff);
  const int minWidth = 32;
  const int minHeight = 32;
  int negHeight = 32;
  int createDirs = 1;
  int skipPositives = 0;
  int skipNegatives = 0;
  
  // read command line parameters
  const int NUM_REQUIRED_PARAMETERS = 2;
  
  const char *imageSequenceFile = NULL;
  const char *objectLabelsFile = NULL;
  const char *negName = "negative";
  const char *posPrefix = NULL;
  const char *negPrefix = NULL;
  
  std::set<std::string> objectNames;
  int resizeWidth = -1;
  int resizeHeight = -1;
  int fponly = 0;
  bool bIncludeFlipped = false;
  bool bIncludeOtherObjects = false;
  int numNegatives = 100;
  double negAspectRatio = 1.0;
  const char *detectionsFilename = NULL;
  float threshold = 0.0;
  const char *baseDirectory = ".";
  svlLogger::setLogLevel(SVL_LOG_MESSAGE);
  
  svlPatchUtil util(SVL_PATCH_CLI_DEPTHCHANNEL);
  bool posBestWindow = false;
  bool allWindows = false;
  int  parallelId = 0;
  int  parallelCount = 0;

  char **args = argv + 1;
  while (--argc > NUM_REQUIRED_PARAMETERS) {
    if (util.parseCommandLineItem(argc, args))
      {
	//Do nothing, util already handled this flag
      } 
      else if (!strcmp(*args, "-object")) {
	objectNames.insert(*(++args));
	argc--;
      } else if (!strcmp(*args, "-aspectRatio")) {
	negAspectRatio = atof(*(++args));
      argc--;
    } 
    else if (!strcmp(*args, "-skipPositives")) {
      skipPositives = 1;
    }
    else if (!strcmp(*args, "-skipNegatives")) {
      skipNegatives = 1;
    }
    else if (!strcmp(*args, "-posBestWindow")) {
      posBestWindow = true; assert(!allWindows);
    }
    else if (!strcmp(*args, "-allWindows")) {
      allWindows = true; assert(!posBestWindow);
    } else if (!strcmp(*args, "-fponly")) {
      fponly = 1;
    }
    else if (!strcmp(*args, "-resize")) {
      resizeWidth = atoi(*(++args));
      resizeHeight = atoi(*(++args));
      argc -= 2;
    } else if (!strcmp(*args, "-negatives")) {
      numNegatives = atoi(*(++args));
      argc--;
    } else if (!strcmp(*args, "-threshold")) {
      threshold = atof(*(++args));
      argc--;
    } else if (!strcmp(*args, "-includeFlipped")) {
      bIncludeFlipped = true;
    } else if (!strcmp(*args, "-includeOtherObjects")) {
      bIncludeOtherObjects = true;
    } else if (!strcmp(*args, "-negHeight")) {
      negHeight = atoi(*(++args));
      argc--;
    }
    else if (!strcmp(*args, "-negName")) 
      {
	negName = *(++args);
	argc--;
      } 
    else if (!strcmp(*args, "-posPrefix")) 
      {
	posPrefix = *(++args);
	argc--;
      } 
    else if (!strcmp(*args, "-negPrefix")) 
      {
	negPrefix = *(++args);
	argc--;
      }
    else if (!strcmp(*args, "-createDirs")) {
      createDirs = 1;
    }
    else if (!strcmp(*args, "-detections")) {
      detectionsFilename = *(++args);
      argc--;
    } else if (!strcmp(*args, "-directory")) {
      baseDirectory = *(++args);
      argc--;
    } else if (!strcmp(*args, "-parallel")) {
      parallelId = atoi(*(++args)); argc--;
      parallelCount = atoi(*(++args)); argc--;
      if (parallelId < 0 || parallelId >= parallelCount) {
        SVL_LOG(SVL_LOG_FATAL, "illegal id number " << parallelId);
      } else if (parallelCount < 0) {
        SVL_LOG(SVL_LOG_FATAL, "illegal job count " << parallelCount);
      }
    }
    else if (!strcmp(*args, "-v")) {
      svlLogger::setLogLevel(SVL_LOG_VERBOSE);
    } 
    
    else {
      SVL_LOG(SVL_LOG_FATAL, "unrecognized option " << *args);
    }
    args++;
  }

  if (argc != NUM_REQUIRED_PARAMETERS) 
    {
      usage();
      return -1;
    }

  
  imageSequenceFile = args[0];
  objectLabelsFile = args[1];


  // load the image sequence 
  SVL_LOG(SVL_LOG_VERBOSE, "Reading image sequence from " << imageSequenceFile << "...");
  svlImageSequence imageSequence;
  imageSequence.load(imageSequenceFile);
  SVL_LOG(SVL_LOG_VERBOSE, "..." << imageSequence.size() << " images read");
  
    //////
  //Complete configuration of the channels
  if (!hasHomogeneousExtensions(imageSequence))
    {
      SVL_LOG(SVL_LOG_FATAL, "Sequence file" << imageSequenceFile << "has multiple file extensions, so no standardized swapping of file extensions will work for managing channels.");
    }

  string baseExt = getExtension(imageSequence);
  util.setBaseExt( baseExt );
  util.finishCommandLine();

  for (unsigned i = 0; i < util.numChannels(); i++)
    {
      if (util.channelTypes(i) == SVL_CHANNEL_EDGE)
	SVL_LOG(SVL_LOG_FATAL, "buildTrainingImageDataset does not export edge patches. Just make them on the fly in later apps.");
    }

  for (unsigned i = 0; i < util.numChannels(); i++)
    {
      channels.push_back(ChannelPair(util.patchTypes(i), util.extensions(i)));
    }

  cout << "Channels: "<< endl;
  for (unsigned int i=0; i < channels.size(); i++)
    {
      switch (channels[i].first)
	{
	case SVL_INTENSITY_PATCH:
	 cout << "INTENSITY";
	 break;
	case SVL_DEPTH_PATCH:
	 cout << "DEPTH:";
	 break;
	case SVL_ABS_PATCH:
	  cout << "ABS:";
	  break;
       default:
	 assert(false);
	}
      
      cout << "\t"<<channels[i].second << endl;
    }


  // load the object labels
  SVL_LOG(SVL_LOG_VERBOSE, "Reading object labels from " << objectLabelsFile << "...");
  svlObject2dSequence objectLabels;
  readObject2dSequence(objectLabelsFile, objectLabels);
  SVL_LOG(SVL_LOG_VERBOSE, "..." << objectLabels.size() << " labeled frames read");
  
  svlObject2dSequence objectDetections;
  
  if (detectionsFilename != NULL) 
    {
      SVL_LOG(SVL_LOG_VERBOSE, "Reading detections from " << detectionsFilename << "...");
      readObject2dSequence(detectionsFilename, objectDetections);
      SVL_LOG(SVL_LOG_VERBOSE, "..." << objectDetections.size() << " labeled frames read");
      int total = countObjects(objectDetections);
      int removed = nonMaximalSuppression(objectDetections, 0.25, 0.25, 0.64);
      SVL_LOG(SVL_LOG_VERBOSE, "..." << removed << " out of " << total << " non-maximal objects suppressed");
    }
  
  // counters for image names
  map<string, int> positiveCounts;
  map<string, int> tempCounts; //since we do each channel sequentially we have to count and then discard counts
  int negativeCounts = 0;
  
  if (createDirs && !fponly && !skipNegatives) {
    string dirName = string(baseDirectory) + string("/") + string(negName);
    svlCreateDirectory(dirName.c_str());
  }


  //List of classifier windows
  unsigned windowListWidth = 0;
  unsigned windowListHeight = 0;
  svlObject2dFrame windows;
  svlSlidingWindowDetector * windowMaker = NULL;
  if (posBestWindow || allWindows)
    {
      windowMaker = new svlPatchBasedObjectDetector("",resizeWidth,resizeHeight);
      assert(windowMaker);
    }

  
  // load images and snip regions
  for (unsigned i = 0; i < imageSequence.size(); i++) 
    {
      if (parallelCount > 0) {
        if ((i % unsigned(parallelCount)) != unsigned(parallelId)) continue ;
      }

      SVL_LOG(SVL_LOG_VERBOSE, "Processing image " << imageSequence[i] << "...");


      string imageId = strBaseName(imageSequence[i]);
      
      //look up the positive examples for this frame
      svlObject2dFrame objects;
      svlObject2dFrame otherObjects;
      if (objectLabels.find(imageId) != objectLabels.end()) 
        {
	  objects = objectLabels[imageId];
	  otherObjects = objectLabels[imageId];
        }
      removeNonMatchingObjects(objects, objectNames);

      if (bIncludeOtherObjects && !objectNames.empty())
	removeMatchingObjects(otherObjects, objectNames);
      else
	otherObjects.clear();

      //make sure there's a running count and a directory for each positive class
      if (!skipPositives)
	{
	  for (unsigned j = 0; j < objects.size(); j++)
	    {
	      if (positiveCounts.find(objects[j].name) == positiveCounts.end()) 
		{
		  positiveCounts[objects[j].name] = 0;
		  
		  if (createDirs) {
		    string dirName = string(baseDirectory) + string("/") +
		      objects[j].name;
		    svlCreateDirectory(dirName.c_str());
		  }                    
		}
	    }

	  tempCounts = positiveCounts;
	}

      //Generate the rectangles to search over for the negatives
      svlObject2dFrame negatives;


      //process each channel of this image
      for (unsigned j = 0; j < channels.size(); j++)
	{

	  //synthesize the filename for this channel
	  string filename = imageSequence.directory() + "/" + imageSequence[i];

	  filename.replace(filename.length()-baseExt.size(), baseExt.size(), channels[j].second);

	  //Load this channel as an IplImage (this works fine for intensity, depth and abs patches, more exotic patches added in the future might want to branch here)
	  IplImage * image = util.loadPatch(filename, channels[j].first);


	  //Check that the channel image was loaded
	  if (!image)
	    {
	      SVL_LOG(SVL_LOG_WARNING, "could not load file " << filename);
	      continue;
	    }

	  //Make sure the windows are up to date
	  if (posBestWindow || allWindows)
	    {
	      
	      if (windowListWidth != (unsigned)image->width || windowListHeight != (unsigned)image->height)
		{
		  windowListWidth = image->width;
		  windowListHeight = image->height;
		  
		  windowMaker->getAllWindows(windowListWidth, windowListHeight, windows);		      
		  
		}		  
	    }
	  
	  if (allWindows)
	    {
	      unsigned negIdx = 0;

	      for (unsigned win = 0; win < windows.size(); win++)
		{
		  bool positive = false;

		  double areaWin = windows[win].area();

		  unsigned obj;

		  for (obj = 0; obj < objects.size(); obj++)
		    {
		      double areaObj = objects[obj].area();
		      double areaOverlap = windows[win].overlap(objects[obj]);

		      if (areaOverlap / (areaObj+areaWin-areaOverlap) > 0.5)
			{
			  positive = true;
			  break;
			}
		    }

		  CvRect r = cvRect((int) windows[win].x, (int) windows[win].y, (int) windows[win].w, (int) windows[win].h);

		  if (positive)
		    {
		      if (!skipPositives)
			{
			  int n = tempCounts[objects[obj].name];

			  handle_clip(image, r, channels[j].first, channels[j].second, baseDirectory, objects[obj].name, posPrefix, objects[obj].name, n, bIncludeFlipped, resizeWidth, resizeHeight);

			  tempCounts[objects[obj].name] = n + 1;
			}
		    }
		  else
		    {
		      if (!skipNegatives)
			{
			  handle_clip(image, r, channels[j].first, channels[j].second, baseDirectory, string(negName), negPrefix, string("image"), negativeCounts + negIdx, false, resizeWidth, resizeHeight);
			  negIdx ++;
			}
		    }
		}

	      //To make sure index numbers agree for all channels we need to reset the counts, unless this is the last channel, in which case we need to store them for the next frame
	      if (j != channels.size() -1)
		tempCounts = positiveCounts; //reset
	      else
		{
		  positiveCounts = tempCounts; //store
		  negativeCounts += negIdx;
		}

	    }
	  else//end allWindow extraction
	    {


	      //Extract positive examples
	      if (!skipPositives)
		{

		  for (unsigned k = 0; k < objects.size(); k++)
		    {
		      CvRect r = cvRect((int)objects[k].x, (int)objects[k].y,(int)objects[k].w, (int)objects[k].h);
		      
		      // check that r does not exceed bounds of image
		      if (r.x+r.width >= image->width)
			r.width = image->width - 1 - r.x;
		      
		      if (r.y+r.height >= image->height)
			r.height = image->height - 1 -r.y;
		      
		      if (posBestWindow)
			{
			  svlObject2d rect(r.x, r.y, r.width, r.height);
			  
			  svlObject2d bestWindow;
			  double bestOverlap = -1;
			  
			  double areaR = rect.area();
			  for (unsigned win = 0; win < windows.size(); win++)
			    {
			      double areaOverlap = rect.overlap(windows[win]);
			      double propOverlap = areaOverlap/(areaR+windows[win].area()-areaOverlap);
			      if (propOverlap > bestOverlap)
				{
				  bestWindow = windows[win];
				  bestOverlap = propOverlap;
				}
			    }
			  assert(bestOverlap > 0);
			  
			  r = cvRect((int)bestWindow.x, (int)bestWindow.y, (int)bestWindow.w, (int)bestWindow.h);
			}
		      
		      
		      int n = tempCounts[objects[k].name];
		      
		      handle_clip(image, r, channels[j].first, channels[j].second, baseDirectory, objects[k].name, posPrefix, objects[k].name, n, bIncludeFlipped, resizeWidth, resizeHeight);
		      
		      
		      tempCounts[objects[k].name] = n + 1;
		    } // end iteration through objects
		  
		  //To make sure index numbers agree for all channels we need to reset the counts, unless this is the last channel, in which case we need to store them for the next frame
		  if (j != channels.size() -1)
		    tempCounts = positiveCounts; //reset
		  else
		    positiveCounts = tempCounts; //store
		} //end extracting positives
	      
	      //Extract negative examples
	      if (!skipNegatives)
		{
		  if (j == 0) //only pick boxes once so that we extract the same patches for each channel. however, we need an image size, so we do it inside the channel loop and assume all channels are the same size
		    getNegatives(negatives, objects, objectDetections, imageId, fponly, numNegatives, negAspectRatio, rngState, image, minWidth, minHeight, negHeight, threshold, otherObjects);
		  
		  for (unsigned int k = 0; k < negatives.size(); k++)
		    {
		      CvRect r = cvRect((int)negatives[k].x, (int)negatives[k].y,(int)negatives[k].w, (int)negatives[k].h);
		      
		      handle_clip(image, r, channels[j].first, channels[j].second, baseDirectory, string(negName), negPrefix, string("image"), negativeCounts + k, false, resizeWidth, resizeHeight);
		    }
		}
	    }//end non-allWindow window extraction
	      
	      cvReleaseImage(&image);

	} //end iterating through channels
      
      negativeCounts += negatives.size();
      
 
    } //end iterating through images

  if (posBestWindow)
    delete windowMaker;
  
  SVL_LOG(SVL_LOG_MESSAGE, "Generated dataset with:");
  
  for (map<string, int>::const_iterator it = positiveCounts.begin();it != positiveCounts.end(); ++it) 
    {
      SVL_LOG(SVL_LOG_MESSAGE, it->second << " instances of " << it->first);
    }
  
  SVL_LOG(SVL_LOG_MESSAGE, negativeCounts << " negative examples");
  
  return 0;
}
  

  
CvRect randRect(double aspectRatio, CvSize base, CvSize frame, CvRNG *rngState)
{
  const double SEL_FACTOR = 100000000.0;
  const double SEL_FACTOR_SQRT = 10000.0;
  int minSel, maxSel;
  CvRect rect;

  // TODO: minSel and maxSel calculation can be done once, given the frame
  // size and base size
  if (frame.width / frame.height > aspectRatio) {
    // Frame is wider than aspect, rectangle is constrained by height
    minSel = (int) ceil(SEL_FACTOR / pow(frame.height, 2.0));
    maxSel = (int) floor(SEL_FACTOR / pow(base.height, 2.0));

    float sel = (float) (cvRandInt(rngState) % (maxSel - minSel) + minSel);
    rect.height = (int) (SEL_FACTOR_SQRT / sqrt(sel));
    rect.width = (int) (rect.height * aspectRatio);
  } else {
    // Frame is thinner than aspect, rectangle is constrained by width
    minSel = (int) ceil(SEL_FACTOR / pow(frame.width, 2.0));
    maxSel = (int) floor(SEL_FACTOR / pow(base.width, 2.0));

    float sel = (float) (cvRandInt(rngState) % (maxSel - minSel) + minSel);
    rect.width = (int) (SEL_FACTOR_SQRT / sqrt(sel));
    rect.height = (int) (rect.width / aspectRatio);
  }

  rect.x = cvRandInt(rngState) % (frame.width - rect.width);
  rect.y = cvRandInt(rngState) % (frame.height - rect.height);

  return rect;
}


void savePatch(const string & filename, const IplImage * patch, svlPatchDefinitionType type)
{
  switch(type)
    {
    case SVL_INTENSITY_PATCH:
    case SVL_ABS_PATCH:
      cvSaveImage(filename.c_str(), patch);
      break;
    case SVL_DEPTH_PATCH:
      writeMatrixAsIplImage(patch, filename.c_str());
      break;
    default:
      assert(false); //unsupported patch type
    }
}


void usage()
{
    cerr << SVL_USAGE_HEADER << endl;
    cerr << "USAGE: ./buildTrainingImageDataset [OPTIONS] <imageSequence> <objectLabels>" << endl;
    cerr << "OPTIONS:" << endl
         << "  -object <name>      :: object extract (default: all)" << endl
         << "  -skipPositives      :: Does not write positive objects to file"<<endl
	 << "  -skipNegatives      :: Does not output negative objects"<<endl
         << "  -negName <name>     :: Name for the folder which contains the negatives (default: negative)"<< endl
         << "  -negHeight <height> :: Used to fix the size of the negative patch in combination with the aspect ratio (default: 32)"<<endl  
         << "  -aspectRatio <num>  :: aspect ratio (w/h) for clipping negative examples (default : 1)" << endl
         << "  -resize <w> <h>     :: resize images to <w>-by-<h>" << endl
         << "  -includeFlipped     :: include horizontally flipped positive examples" << endl
         << "  -negatives <n>      :: number of negatives per image (default: 100)" << endl
	 << "  -includeOthers      :: will specifically extract all other labeled objects as negatives first, before doing random selection" << endl
         << "  -fponly             :: processes only the false positives from the given detections and does not add any new negatives (default: inactive)" << endl
         << "  -detections <file>  :: include false-positive detections from <file>" << endl
	 << "  -threshold <decimal>:: include only the false-positive detections that exceed this probability threshold (default: 0)" << endl
         << "  -createDirs         :: Creates sub directories for object names & one for negative images" << endl
         << "  -directory <dir>    :: output base directory " << endl
         << "  -posPrefix <string> :: prefix for the final names of the positive images (default : blank)" << endl
         << "  -negPrefix <string> :: prefix for the final names of the negative images (default : blank)" << endl
	 << "  -dmsize <w> <h>     :: dimensions of the depth map"<< endl
         << "  -v                  :: verbose (default: false)" << endl
         << "  -ch <type> <ext>    :: adds a channel. type is INTENSITY, DEPTH, or ABS. ext is - to use extension from image sequence, otherwise it is your own string (ie, \".img.jpg\" or \".flow.bmp\"). if no channels are specified INTESNITY - is added by default" << endl
         << " -baseExt <ext>       :: filename extension to be replaced when patching in the different extension for each channel. also used as the default value for channels whose extension is \"-\":. default is to query the image sequence for this value. " << endl
	 << " -posBestWindow       :: instead of using ground truth labels as positive windows, use the best-matching window from the sliding window detector" << endl
	 << " -allWindows          :: extracts ALL windows, saving them as positive if they have 50% or greater overlap with a ground truth object." << endl
	 << " -parallel <id> <ct>  :: used for distributed jobs on the cluster." << endl
         << "DEPRECATED OPTIONS: " << endl
         << "  -imageExt <ext>     :: extension of the image data"<< endl
         << "  -depthMapExt <ext>  :: extension of the depth map data"<< endl
         << "  -depthChannel       :: indicates the presence of a depth channel for all images in the image seq"<<endl
	 << endl;
}

//todo-- use of typeNameForFile is kind of stupid. I (IG) only put it there
//for consistency with the way the app used to to behave before I put this
//into a single function. This all used to be independently hardcoded for
//positive and negative examples, and for negative it would use negName
//in one place but "image" in another, and for positive it would use
//the object name in both places
void handle_clip(IplImage * image, const CvRect & r, svlPatchDefinitionType pType, const string & ext, string baseDirectory, string typeName, const char * prefix, string typeNameForFile, unsigned int idx, bool bIncludeFlipped, int resizeWidth, int resizeHeight)
 {
   IplImage * clip = NULL;
   
   cvSetImageROI(image, r);

   CvRect rect = cvGetImageROI(image);

   //patchtype specific format for image patch to clip out
   switch (pType)
     {
     case SVL_INTENSITY_PATCH:
       clip = cvCreateImage(cvSize(rect.width, rect.height), IPL_DEPTH_8U, 1);
       break;
     case SVL_ABS_PATCH:
       clip = cvCreateImage(cvSize(rect.width, rect.height), IPL_DEPTH_8U, 3);
       break;
     case SVL_DEPTH_PATCH:
       clip = cvCreateImage(cvSize(rect.width, rect.height), IPL_DEPTH_32F, 1);
       break;		   
     default:
       assert(false); //unsupported or invalid patch type
     }

   assert(clip);

   cvCopy(image, clip);
  
  cvResetImageROI(image);
  
  if (resizeWidth > 0 && resizeHeight > 0)
    resizeInPlace(&clip, resizeHeight, resizeWidth);
  
  //patchtype specific post-processing
  switch (pType)
    {
    case SVL_INTENSITY_PATCH:
    case SVL_ABS_PATCH:
      //No post-processing needed
      break; 
    case SVL_DEPTH_PATCH:
      svlSubtractMedian(clip);
      break;
    default:
      assert(false);
    }

  string baseFilename = string(baseDirectory) + string("/") + typeName +string("/");
  if (prefix)
    baseFilename += prefix;
  baseFilename += typeNameForFile;
  baseFilename += padString(toString(idx), 5);

  string filename = baseFilename + ext;

  savePatch(filename, clip, pType);

  if (bIncludeFlipped)
    {
      cvFlip(clip, NULL, 1);
      
      string flippedFilename = baseFilename + "f" + ext;
      
      savePatch(flippedFilename.c_str(), clip, pType);
    }

  cvReleaseImage(&clip);
}

void getNegatives(svlObject2dFrame & output, const svlObject2dFrame & objects, svlObject2dSequence & objectDetections,
		  const string & imageId, const int & fponly, const int & numNegatives, const double & negAspectRatio,
		  CvRNG & rngState, IplImage * image, const int & minWidth, const int & minHeight,
		  const int & negHeight, const float &threshold, const svlObject2dFrame &otherObjects)
{
  // extract negative examples
  svlObject2dFrame detections;

  if (objectDetections.find(imageId) != objectDetections.end())
    {
      detections = objectDetections[imageId];
      removeGroundTruthObjects(detections, objects, 0.5);
      detections = sortObjects(detections);
    }

  //add worst false-positives, if those exist
  for (unsigned j = 0; j < (unsigned)numNegatives; j++)
    {
      if (j >= detections.size() || detections[j].pr < threshold)
	break;
      
      CvRect r = cvRect((int)detections[j].x, (int)detections[j].y, (int) detections[j].w, (int) detections[j].h);
      
      output.push_back(r);
    }


  // add other negatives if necessary
  if (!fponly)
    {
      // first, add the other objects as negative examples; then will add random rectangles as necessary

      int numNegsLeft = numNegatives - output.size();
      int numOtherObjectsToAdd = min((int)otherObjects.size(), numNegsLeft);
      if (numOtherObjectsToAdd > 0) {
	output.insert(output.end(), otherObjects.begin(), otherObjects.begin() + numOtherObjectsToAdd);
	numNegsLeft = numNegatives - output.size();
      }

      for (int j = 0; j < numNegsLeft; j++)
	{
	  CvRect r;

	  //Generate random rectangle
	  if (negAspectRatio <= 0.0)
	    {
	      r.width = (cvRandInt(&rngState) % (image->width - minWidth)) + minWidth;
	      r.height = (cvRandInt(&rngState) % (image->height - minHeight)) + minHeight;
	    }
	  else if (negAspectRatio > 1.0)
	    {
	      if (negHeight == -1)
		{
		  r.width = (cvRandInt(&rngState) % (image->width - minWidth)) + minWidth;
		  r.height = (int)(r.width / negAspectRatio);
		}
	      else
		{
		  r.height = negHeight;
		  r.width = (int)(negAspectRatio * r.height);
		}
	    }
	  else
	    {
	      if (negHeight == -1)
		{
		  r.height = (cvRandInt(&rngState) % (image->height - minHeight)) + minHeight;
		  r.width = (int)(negAspectRatio * r.height);
		}
	      else
		{
		  r.height = negHeight;
		  r.width = (int)(negAspectRatio * r.height);
		}
	    }

	 //************************** Check for overlap with all ground truth objects in image ************
	 //*********************************** and for uniformity of the patch ****************************

	  bool found;
	  int numTries = 0;

	  do
	    {
	      found = true;

	      r.x = cvRandInt(&rngState) % (image->width - r.width);
	      r.y = cvRandInt(&rngState) % (image->height - r.height);

	      svlObject2d temp(r.x,r.y,r.width,r.height);

	      // if the image is almost uniform, don't use it, found = false
	      cvSetImageROI(image, r);
	      if (isImageAlmostUniform(image)) found = false;
	      cvResetImageROI(image);

	      // search through all the objects (as long as you still believe you found a good location)
	      for (unsigned k =0; (k < objects.size()) && found; k++)
		{
		  if (objects[k].overlap(temp) != 0)
		      found = false;

		}

	      // it's possible that there's just no good position
	      // to put it; in that case, continue
	      numTries++;
	    } while (!found && numTries < MAX_NUM_TRIES);

	  if (found)
	    output.push_back(r);
	}
    }
}
