/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlGeneralObjectDetector.cpp
** AUTHOR(S):   Ian Goodfellow <ia3n@cs.stanford.edu>
**             
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <limits>

#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlVision.h"

using namespace std;


static void markUsedVars(CvDTreeNode * n, vector<bool> & used)
{
  if (n->left)
    markUsedVars(n->left, used);

  if (n->right)
    markUsedVars(n->right, used);


  if (n->split)
    {
      used.at(n->split->var_idx) = true;
      assert(!n->split->next);
    }
}

static void markUsedVars(CvBoostTree * t,  vector<bool> & used)
{
  if (t->get_root())
    markUsedVars((CvDTreeNode *)t->get_root(), used);
}

static void mapTree(CvDTreeNode * n, const vector<unsigned> & map)
{
  if (n->left)
    mapTree(n->left, map);

  if (n->right)
    mapTree(n->right, map);


  if (n->split)
    {
      n->split->var_idx = map.at(n->split->var_idx);
      assert(!n->split->next);
    }
}

static void mapTree(CvBoostTree * t, const vector<unsigned> & map)
{
  if (t->get_root())
    mapTree((CvDTreeNode *) t->get_root(), map);
}



// svlGeneralObjectDetector ----------------------------------------------

svlGeneralObjectDetector::svlGeneralObjectDetector(const char *name, int windowWidth, int windowHeight) :
  svlSlidingWindowDetector(name, windowWidth, windowHeight), _extractor(NULL)
{
    // do nothing
}

svlGeneralObjectDetector::svlGeneralObjectDetector(const svlGeneralObjectDetector& c) :
  svlSlidingWindowDetector(c), _extractor(c._extractor->clone())
{
    // do nothing
}

svlGeneralObjectDetector::~svlGeneralObjectDetector()
{
  if (_extractor)
    delete _extractor;
}

// configuration and access functions
bool svlGeneralObjectDetector::readFeatures(const char *filename)
{
  if (_extractor)
    delete _extractor;

  _extractor = svlDetectorFeatureExtractor::getInstance(filename);

  //cout << "Read extractor with "<<_extractor->numFeatures()<<" features"<<endl;

    if (_extractor) {
        _windowWidth = _extractor->windowWidth();
        _windowHeight = _extractor->windowHeight();
        return true;
    }

    return false;
}
 
bool svlGeneralObjectDetector::writeFeatures(const char *filename)
{
    return _extractor->write(filename);
}



void svlGeneralObjectDetector::computeImageFeatureVectors(vector<vector<double> >& fv,
							     const vector<CvPoint>& locations,
							     const vector<IplImage *>& images, bool sparse)
{
  assert(_extractor);
  _extractor->extract(locations, images, fv,  sparse);
}

void svlGeneralObjectDetector::computePatchFeatureVector(vector<double> &fv,
    const vector<IplImage *>& images)
{
  _extractor->windowResponse(images, fv);	
}


void svlGeneralObjectDetector::trim() 
{
  assert(_extractor);
  assert(_model);


  //make the CvSeq of weak learners into a
  //more palatable stl vector

  assert(false); //this doesn't actually work yet because there's no way to
  //update _model->data

  /*
  if (_model->data->var_idx)
    {
      cout << "var_idx exists" << endl;
    }
  else
    {
      cout << "var_idx does not exist" << endl;
    }
  */

  CvSeq * weak = _model->get_weak_predictors();
  //this is a reference to the actual weak learners, not a copy, so modifying
  //them should modify model

  vector<CvBoostTree *> trees;

  CvSeqReader reader;

  cvStartReadSeq(weak, &reader);

  for (int i = 0; i < weak->total; i++)
    {
      CvBoostTree * t;
      CV_READ_SEQ_ELEM( t, reader);

      trees.push_back(t);
    }

  
 

 //determine which features are used
  vector<bool> used(_extractor->numFeatures(), false);

  cout << _extractor->numFeatures() << endl;

  for (unsigned i = 0; i < trees.size(); i++)
    {
      markUsedVars(trees[i],used);
    }


  //trim extractor
  svlDetectorFeatureExtractor * trimmed = _extractor->getPrunedExtractor(used);
  delete _extractor;
  _extractor = trimmed;

  //use it to trim model
  vector<unsigned> map(used.size());

  unsigned newIdx = 0;
  for (unsigned i = 0; i < map.size(); i++)
    {
      if (used[i])
	{
	  map[i] = newIdx;
	  newIdx ++;
	}
    }

  for (unsigned i = 0; i < trees.size(); i++)
    mapTree(trees[i], map);



}
