/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPlane.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**   Class for encapsulating two svlPoint3d's for clustering centroids.
*****************************************************************************/

#include "svlPlane.h"

svlPlane::svlPlane() {
}

svlPlane::svlPlane(double val) : p(svlPoint3d(val)), n(svlPoint3d(val)) {
	off = 3.0*val*val;
}

svlPlane::svlPlane(double p0, double p1, double p2, double n0, double n1, double n2) 
	: p(svlPoint3d(p0,p1,p2)), n(svlPoint3d(n0,n1,n2)) {
	off = p.dot(n);
}

svlPlane::svlPlane(const svlPoint3d& p, const svlPoint3d& n) : p(svlPoint3d(p)), n(svlPoint3d(n)) {
	off = p.dot(n);
}

svlPlane::~svlPlane() {
	// do nothing
}

svlPlane& svlPlane::set(double val) {
	p.set(val);
	n.set(val);
	off = val;
	return *this;
}

svlPlane& svlPlane::set(const svlPlane& pl) {
	p = pl.p;
	n = pl.n;
	off = pl.off;
	return *this;
}

// svlPlane& svlPlane::set(const svlPlane& c_in) {
//	p = c_in.p;
//	n = c_in.n;
//	return *this;
//}
//
// svlPlane& svlPlane::set(const svlPoint3d& p_in, const svlPoint3d& n_in) {
//	p = p_in;
//	n = n_in;
//	return *this;
//}
//
// svlPlane& svlPlane::add(const svlPoint3d& p_in, const svlPoint3d& n_in) {
//	p += p_in;
//	n += n_in;
//	return *this;
//}

double svlPlane::norm() const {
    return fabs(off);
}

double svlPlane::norm2() const {
    return off*off;
}

double svlPlane::perp_distance(const svlPoint3d &pt) const {
    return n.dot(pt)-off;
}

double svlPlane::dist(const svlPoint3d &pt) const {
    double d = fabs(n.dot(pt) - off);
    
    // check if point it within plane
    svlPoint2d pt2 = project(pt);
    if (!((pt2 >= min_pt2) && (pt2 <= max_pt2))) {
	// compute distance to nearest point in bounded plane area
	double x, y;
	x = ( pt2.x >= max_pt2.x ? pt2.x-max_pt2.x
	    : ( pt2.x < min_pt2.x ? pt2.x-min_pt2.x
		: 0.0 ) );
	y = ( pt2.y >= max_pt2.y ? pt2.y-max_pt2.y
	    : ( pt2.y < min_pt2.y ? pt2.y-min_pt2.y
		: 0.0 ) );
	d = sqrt(d*d + x*x + y*y);
    }

    return d;
}

void svlPlane::offset(const svlPoint3d &t) {
	p += t;
	off = n.dot(p);
	for ( vector<svlPoint3d>::iterator it=Hull3d.begin(); it!=Hull3d.end(); ++it ) {
	    *it += t;
	}
}

void svlPlane::rotateX(double theta) {
    p = p.tilt(theta);
    n = n.tilt(theta);
    u1 = u1.tilt(theta);
    u2 = u2.tilt(theta);
    for ( vector<svlPoint3d>::iterator it=Hull3d.begin(); it!=Hull3d.end(); ++it ) {
	*it = it->tilt(theta);
    }
    for ( vector<svlPoint3d>::iterator it=HullPs.begin(); it!=HullPs.end(); ++it ) {
	*it = it->tilt(theta);
    }
}

svlPlane& svlPlane::operator/=(const double d) {
	p /= d;
	off /= d;
	return *this;
}

svlPoint3d operator-(const svlPlane& pl1, const svlPlane& pl2) {
	return pl1.p - pl2.p; // displacement vector between both planes' centroid vectors
}

bool svlPlane::isVertical() {
	//assert( (n.norm2()-1.0) <= 1e-6) ); // assert normal vector is normalized (heh)
	return fabs(n[1]) <= 0.08; // very small y component
}

bool svlPlane::isHorizontal() {
	//assert( (n.norm2()-1.0) <= 1e-6) ); // assert normal vector is normalized (heh)
	return fabs(n[1]) >= 0.92; // dominant y component
}

svlPoint2d svlPlane::project(const svlPoint3d &pt) const {
    return svlPoint2d(u1.dot(pt-p), u2.dot(pt-p));
}

svlPoint3d svlPlane::project3d(const svlPoint3d &pt) const {
	return u1.dot(pt-p)*u1 + u2.dot(pt-p)*u2;
}

svlPoint3d svlPlane::intercept3d(svlPoint3d ray) const {
	return (n.dot(p)/n.dot(ray)) * ray;
}

double svlPlane::interceptScale(svlPoint3d ray) const {
	return n.dot(p)/n.dot(ray);
}

svlPoint3d& svlPlane::interceptThis(svlPoint3d &ray) {
	return ( ray *= n.dot(p)/n.dot(ray) );
}

double svlPlane::minAbsPerpDist(const vector<svlPoint3d> &P) {
	vector<svlPoint3d>::const_iterator it = P.begin();
#define MAPD_DEBUG_OUT 0
#if MAPD_DEBUG_OUT
	static int ctr;
	char c[32];
	sprintf(c,"zzz_mapd_%d.out",ctr);
	FILE *fo = fopen(c,"wb");
	fwrite(&n.x,sizeof(double),1,fo);
	fwrite(&n.y,sizeof(double),1,fo);
	fwrite(&n.z,sizeof(double),1,fo);
	fwrite(&off,sizeof(double),1,fo);
#endif
	double gd = fabs(n.dot(*it)-off);
#if MAPD_DEBUG_OUT
	fwrite(&gd,sizeof(double),1,fo);
#endif
	for ( ++it; it!=P.end(); ++it ) {
		gd = gd < fabs(n.dot(*it)-off) ? gd : fabs(n.dot(*it)-off);
#if MAPD_DEBUG_OUT
		double gd1 = fabs(n.dot(*it)-off);
		fwrite(&gd1,sizeof(double),1,fo);
#endif
	}
#if MAPD_DEBUG_OUT
	fclose(fo);
#endif
	return gd;
}

void svlPlane::heightMeanVar(const vector<svlPoint3d> &P, double &mean, double &var) {
	mean = var = 0.0;
	if ( (int)P.size() == 0 ) return;

	double gd;
	for ( vector<svlPoint3d>::const_iterator i=P.begin(); i!=P.end(); ++i ) {
		gd = perp_distance(*i);
		mean += gd;
		var += gd*gd;
	}
	mean /= (double)P.size();
	var = var/(double)P.size() - mean*mean;
}

void svlPlane::heightMeanVar(const vector<svlPoint3d> &P, const vector<int> &ix, double &mean, double &var) {
	mean = var = 0.0;
	if ( (int)P.size() == 0 ) return;

	double gd;
	for ( vector<int>::const_iterator i=ix.begin(); i!=ix.end(); ++i ) {
		gd = perp_distance(P[*i]);
		mean += gd;
		var += gd*gd;
	}
	mean /= (double)P.size();
	var = var/(double)P.size() - mean*mean;
}


bool svlPlane::supports(svlPoint3d pt, double &dist) {
	dist = perp_distance(pt); // signed distance from plane
	svlPoint2d pt2 = project(pt);
	return pt2 >= min_pt2 && pt2 <= max_pt2;
}

bool svlPlane::supports(svlPoint3d pt, double &dist, double &indist) {
	dist = n.dot(pt) - off; // signed distance from plane
	svlPoint2d pt2 = project(pt);
	bool gb = pt2 >= min_pt2 && pt2 <= max_pt2;
	if ( gb ) {
		indist = std::min( (pt2-min_pt2).pt_min(), (max_pt2-pt2).pt_min() );
	} else {
		// compute distance to nearest point in bounded plane area
		double x, y;
		x = ( pt2.x >= max_pt2.x ? pt2.x-max_pt2.x
			: ( pt2.x < min_pt2.x ? pt2.x-min_pt2.x
				: 0.0 ) );
		y = ( pt2.y >= max_pt2.y ? pt2.y-max_pt2.y
			: ( pt2.y < min_pt2.y ? pt2.y-min_pt2.y
				: 0.0 ) );
		indist = sqrt(x*x + y*y);
	}
	return gb;
}

double svlPlane::percentSupport(vector<svlPoint3d> &P) {
#define RAB_DEBUG_OUT 0
	// project the points into the plane, compute singular vector bounds, perform math
	svlPoint2d in_min_pt2(1e6), in_max_pt2(-1e6), pt2;
	for ( vector<svlPoint3d>::iterator i=P.begin(); i!=P.end(); ++i ) {	    
		pt2 = project(*i);
		in_min_pt2.min_with(pt2);
		in_max_pt2.max_with(pt2);
	}
	double x_perc = 0.0, y_perc = 0.0;

	if (in_max_pt2.x == in_min_pt2.x) return 0.0;
	if (in_max_pt2.y == in_min_pt2.y) return 0.0;

	// compute percentage overlap (w.r.t. the largest object (the plane or the input box)) in each dimension; multiply for area
#if 0
	x_perc = ( std::min(in_max_pt2.x, max_pt2.x) - std::max(in_min_pt2.x, min_pt2.x) ) / (max_pt2.x-min_pt2.x);
	y_perc = ( std::min(in_max_pt2.y, max_pt2.y) - std::max(in_min_pt2.y, min_pt2.y) ) / (max_pt2.y-min_pt2.y);
#elif 0
	x_perc = ( std::min(in_max_pt2.x, max_pt2.x) - std::max(in_min_pt2.x, min_pt2.x) ) / std::max(max_pt2.x-min_pt2.x, in_max_pt2.x-in_min_pt2.x);
	y_perc = ( std::min(in_max_pt2.y, max_pt2.y) - std::max(in_min_pt2.y, min_pt2.y) ) / std::max(max_pt2.y-min_pt2.y, in_max_pt2.y-in_min_pt2.y);
#else
	x_perc = ( std::min(in_max_pt2.x, max_pt2.x) - std::max(in_min_pt2.x, min_pt2.x) ) / (in_max_pt2.x-in_min_pt2.x);
	y_perc = ( std::min(in_max_pt2.y, max_pt2.y) - std::max(in_min_pt2.y, min_pt2.y) ) / (in_max_pt2.y-in_min_pt2.y);
#endif

#if RAB_DEBUG_OUT
	static int percind;
	char c[32];
	sprintf(c,"vvv_perc_%d.out",percind++);
	FILE *fo = fopen(c,"wb");
	fwrite(&in_min_pt2.x,sizeof(double),1,fo);
	fwrite(&in_min_pt2.y,sizeof(double),1,fo);
	fwrite(&in_max_pt2.x,sizeof(double),1,fo);
	fwrite(&in_max_pt2.y,sizeof(double),1,fo);
	fwrite(&x_perc,sizeof(double),1,fo);
	fwrite(&y_perc,sizeof(double),1,fo);
	fclose(fo);
#endif

	if ( x_perc < 0.0 || y_perc < 0.0 ) return 0.0;
	x_perc *= y_perc;
	if ( x_perc > 1.0 ) x_perc = 1.0;
	return x_perc;
}

double svlPlane::percentSupport(vector<svlPoint3d> &P, vector<int> &ix) {
	// project the points into the plane, compute singular vector bounds, perform math
	svlPoint2d in_min_pt2(1e6), in_max_pt2(-1e6), pt2;
	for ( vector<int>::const_iterator i=ix.begin(); i!=ix.end(); ++i ) {
		pt2 = project(P[*i]);
		in_min_pt2.min_with(pt2);
		in_max_pt2.max_with(pt2);
	}
	double x_perc = 0.0, y_perc = 0.0;
	// compute percentage overlap (w.r.t. the largest object (the plane or the input box)) in each dimension; multiply for area
	x_perc = ( std::min(in_max_pt2.x, max_pt2.x) - std::max(in_min_pt2.x, min_pt2.x) ) / std::max(max_pt2.x-min_pt2.x, in_max_pt2.x-in_min_pt2.x);
	y_perc = ( std::min(in_max_pt2.y, max_pt2.y) - std::max(in_min_pt2.y, min_pt2.y) ) / std::max(max_pt2.y-min_pt2.y, in_max_pt2.y-in_min_pt2.y);
	if ( x_perc < 0.0 || y_perc < 0.0 ) return 0.0;
	return x_perc * y_perc;
}

void svlPlane::computeHulls(const vector<svlPoint3d> &Pts, const vector<int> &ix) {
	CvMat *P = cvCreateMat(3,(int)Pts.size(),CV_32FC1);
	CvMat *P2 = cvCreateMat(2,P->cols,CV_32FC1);
	CvMat *PPT = cvCreateMat(3,3,CV_32FC1);
	CvMat *U = cvCreateMat(3,3,CV_32FC1); // left orthogonal matrix
	CvMat *S = cvCreateMat(3,3,CV_32FC1); // singular values
	CvMat *l = cvCreateMat(3,1,CV_32FC1); // vector
	CvMat *U12 = cvCreateMat(2,3,CV_32FC1); // projection matrix
	CvMat *minmax = cvCreateMat(2,2,CV_32FC1);
	CvMemStorage *storage = cvCreateMemStorage(100);

	computeHulls(Pts, ix, P, P2, PPT, U, S, l, U12, minmax, storage);

	cvReleaseMat(&P);
	cvReleaseMat(&P2);
	cvReleaseMat(&PPT);
	cvReleaseMat(&U);
	cvReleaseMat(&S);
	cvReleaseMat(&l);
	cvReleaseMat(&U12);
	cvReleaseMat(&minmax);
	cvReleaseMemStorage(&storage);
}

void svlPlane::computeHulls(const vector<svlPoint3d> &Pts, const vector<int> &ix,
							CvMat* P, CvMat* P2, CvMat* PPT, CvMat* U, CvMat* S, CvMat* l, CvMat* U12, CvMat* minmax, CvMemStorage* storage) {
#define PCH_DEBUG_OUT 0
#if PCH_DEBUG_OUT
	FILE *fo; int foi;
#endif
	if ( P->cols < (int)ix.size() || P->cols >= (((int)ix.size())<<1) ) {
		// too little size or too much by a factor of 2
		//cvReleaseMat(&P);
		//cvReleaseMat(&P2);
		P = cvCreateMat(3,(int)ix.size(),CV_32FC1);
		P2 = cvCreateMat(2,(int)ix.size(),CV_32FC1);
	}

	vector<svlPoint2d> subP;
	svlPoint3d pt3;

	// compute mean of points
	p = 0.0;
	n = 0.0; // zeralize plane parameters
#if PCH_DEBUG_OUT
	fo = fopen("zzz_h_1.out","wb");
	fwrite(&(foi=(int)ix.size()),sizeof(int),1,fo);
	fwrite(&(foi=(int)Pts.size()),sizeof(int),1,fo);
	fclose(fo);
#endif
	for ( vector<int>::const_iterator m=ix.begin(); m!=ix.end(); ++m ) {
		p += Pts[*m];
	}
	p /= (double)ix.size();
#if PCH_DEBUG_OUT
	fwrite(&(foi=(int)ix.size()),sizeof(int),1,fo);
	fwrite(&(foi=P->cols),sizeof(int),1,fo);
	fclose(fo);
#endif

	// assemble P matrix
	int k = 0;
	for ( ; k<(int)ix.size(); ++k ) {
		pt3 = Pts[ix[k]] - p;
		CV_MAT_ELEM(*P,float,0,k) = (float)pt3.x;
		CV_MAT_ELEM(*P,float,1,k) = (float)pt3.y;
		CV_MAT_ELEM(*P,float,2,k) = (float)pt3.z;
	}
	// zeralize remaining columns
	for ( ; k<P->cols; ++k ) {
		CV_MAT_ELEM(*P,float,0,k) = 0.0;
		CV_MAT_ELEM(*P,float,1,k) = 0.0;
		CV_MAT_ELEM(*P,float,2,k) = 0.0;
	}

	// compute P*P^T, SVD
	cvGEMM(P,P,1.0,NULL,0.0,PPT,CV_GEMM_B_T); // P*P^T -> PPT
	cvSVD(PPT,S,U,NULL,0); // SVD to get principal vectors
	// write U1 and U2 into U12 manually
	CV_MAT_ELEM(*U12,float,0,0) = CV_MAT_ELEM(*U,float,0,0);
	CV_MAT_ELEM(*U12,float,0,1) = CV_MAT_ELEM(*U,float,1,0);
	CV_MAT_ELEM(*U12,float,0,2) = CV_MAT_ELEM(*U,float,2,0);
	CV_MAT_ELEM(*U12,float,1,0) = CV_MAT_ELEM(*U,float,0,1);
	CV_MAT_ELEM(*U12,float,1,1) = CV_MAT_ELEM(*U,float,1,1);
	CV_MAT_ELEM(*U12,float,1,2) = CV_MAT_ELEM(*U,float,2,1);
	// save svlPoint3d versions of U analyzing vectors
	u1 = svlPoint3d(CV_MAT_ELEM(*U,float,0,0),
					CV_MAT_ELEM(*U,float,1,0),
					CV_MAT_ELEM(*U,float,2,0));
	u2 = svlPoint3d(CV_MAT_ELEM(*U,float,0,1),
					CV_MAT_ELEM(*U,float,1,1),
					CV_MAT_ELEM(*U,float,2,1));
	n = svlPoint3d(CV_MAT_ELEM(*U,float,0,2),
				   CV_MAT_ELEM(*U,float,1,2),
				   CV_MAT_ELEM(*U,float,2,2));
	off = n.dot(p);
	cvGEMM(U12,P,1.0,NULL,0.0,P2,0); // compute 2D projection: U12*P -> P2

	// data structure prep for convex hull
	CvSeq* ptseq = cvCreateSeq( CV_SEQ_ELTYPE_POINT|CV_32FC1, sizeof(CvContour), sizeof(CvPoint), storage);
	CvPoint2D32f pt0;
	// TOMIGHT [PB]: Hull straight to matrix to speed transition back to 3D pts
    for( int j=0; j<(int)ix.size(); ++j ) { // push points onto sequence just to waste time
		pt0.x = CV_MAT_ELEM(*P2,float,0,j);
        pt0.y = CV_MAT_ELEM(*P2,float,1,j);
        cvSeqPush(ptseq,&pt0);
	}
	// comput convex hull
	CvSeq* hull = cvConvexHull2(ptseq,0,CV_CLOCKWISE,1); // 1 for get points back

	// save 2D and 3D hulls and 2D bounds
	subP.resize(hull->total);
	Hull2d.resize(hull->total); // zeros for 2D hull
	min_pt2.set(1e9);
	max_pt2.set(-1e9);
	Hull3d = vector<svlPoint3d>(hull->total,p); // hull->total copies of the 3D centroid vector
	// convert hull straight to svlPoint3d's: no matrix multiplication
	for( int i=0; i<hull->total; i++ ) {
        pt0 = *CV_GET_SEQ_ELEM(CvPoint2D32f, hull, i);
		Hull2d[i] = svlPoint2d(pt0.x,pt0.y);
		min_pt2.min_with(Hull2d[i]);
		max_pt2.max_with(Hull2d[i]);
		Hull3d[i] += pt0.x*u1 + pt0.y*u2;
    }

	svlPoint3d n1, n2;
	vector<svlPoint3d>::const_iterator it1 = Hull3d.begin(), it2, it3;
	
	// compute hull bounds
	hullLims.clear();
	HullPs.clear(); // normals to each segment of the hull in 3d
	hullLims.reserve(Hull3d.size());
	HullPs.reserve(Hull3d.size());
	double gd, gdmax;
	svlPoint3d plt;
	for ( it1=Hull3d.begin(); it1!=Hull3d.end(); ++it1 ) {
		it2 = it1 + 1;
		if ( it2 == Hull3d.end() ) it2 = Hull3d.begin();
		HullPs.push_back(p-*it1);
		plt = (*it2-*it1).normalize();
		HullPs.back() -= HullPs.back().dot(plt)*plt;
		HullPs.back().normalize();
		gdmax = HullPs.back().dot(*it1);
		for ( vector<svlPoint3d>::const_iterator it3=Hull3d.begin(); it3!=Hull3d.end(); ++it3 )
			if ( (gd=HullPs.back().dot(*it3-*it1)) > gdmax )
				gdmax = gd;
		hullLims.push_back(gdmax);
	}

	/*FILE *fo = fopen("zzz_pl.out","wb");
	fwrite(&p.x,sizeof(double),1,fo);
	fwrite(&p.y,sizeof(double),1,fo);
	fwrite(&p.z,sizeof(double),1,fo);
	for ( int i=0; i<(int)Hull3d.size(); ++i ) {
		fwrite(&Hull3d[i].x,sizeof(double),1,fo);
		fwrite(&Hull3d[i].y,sizeof(double),1,fo);
		fwrite(&Hull3d[i].z,sizeof(double),1,fo);
	}
	for ( int i=0; i<(int)Hull3d.size(); ++i ) {
		fwrite(&HullPs[i].x,sizeof(double),1,fo);
		fwrite(&HullPs[i].y,sizeof(double),1,fo);
		fwrite(&HullPs[i].z,sizeof(double),1,fo);
	}
	for ( int i=0; i<(int)Hull3d.size(); ++i ) {
		fwrite(&hullLims[i],sizeof(double),1,fo);
	}
	fclose(fo);*/

#if PCH_DEBUG_OUT
	fo = fopen("zzz_h_5.out","wb"); fclose(fo);
#endif
}

bool svlPlane::insideHull(svlPoint3d pt) {
	bool ret = true;
	double gd;
#define INSIDE_TOL 1e-6
	for ( int i=0; ret && i<(int)HullPs.size(); ++i )
		ret = (gd=HullPs[i].dot(pt-Hull3d[i])) >= -INSIDE_TOL && gd <= hullLims[i]+INSIDE_TOL;
	return ret;
}

void svlPlane::trimHull() {
	trimHull(20); // max of twenty points on hull
}

void svlPlane::trimHull(int h_size) {
	if ( (int)Hull2d.size() > h_size ) { // trim back the hull
		int n = (int)Hull2d.size()-1; // last valid index
		int jm, jp, gi; // we'll need these
		svlSortVectorDI max_sep(true); // sort descending
		vector<svlPoint2d>::iterator it2;
		vector<svlPoint3d>::iterator it3;
		max_sep.reserve(n+1);
		// push on first distances
		max_sep.push_back( std::max( (Hull2d[0]-Hull2d[n]).norm2(), (Hull2d[1]-Hull2d[0]).norm2() ) );
		for ( int i=1; i<=n-1; ++i ) {
			max_sep.push_back( std::max( (Hull2d[i]-Hull2d[i-1]).norm2(), (Hull2d[i]-Hull2d[i+1]).norm2() ) );
		}
		max_sep.push_back( std::max( (Hull2d[n]-Hull2d[n-1]).norm2(), (Hull2d[0]-Hull2d[n]).norm2() ) );

		while ( (int)max_sep.size() > h_size ) {
			/*cout << "The size: " << max_sep.size() << ":\n";
			for ( vector<pair<double,int> >::iterator i=max_sep.myvec.begin(); i!=max_sep.myvec.end(); ++i ) {
				cout << i->first << ": " << i->second << endl;
			}*/

			max_sep.sort();
			int j = max_sep.back().second; // first element is point to kill
			//cout << "Kicking out: " << j << " __ " << max_sep.back().first << ": " << max_sep.back().second << endl;
			// recalc for left-side and right-side points
			for ( jm=j-1; jm!=j; --jm ) {
				if ( jm < 0 ) jm = n; // loop around
				if ( (gi=max_sep.get_orig_ix(jm)) != -1 ) { // this guy's still in
					break;
				}
			}
			for ( jp=j+1; jm!=j; ++jm ) {
				if ( jp > n ) jp = 0; // loop around
				if ( (gi=max_sep.get_orig_ix(jp)) != -1 ) { // this guy's still in
					break;
				}
			}
			//cout << "jm: " << jm << "; jp: " << jp << endl;
			// jm and jp are the indices of the points that used to bound j; recalc these points' distances
			for ( gi=jm-1; gi!=jp; --gi ) {
				if ( gi < 0 ) gi = n; // loop around
				if ( max_sep.get_orig_ix(gi) != -1 ) { // this guy's still in
					break;
				}
			}
			//cout << "gi: " << gi << "; jm: " << jm << endl;
			max_sep[max_sep.get_orig_ix(jm)].first = std::max( (Hull2d[jm]-Hull2d[jp]).norm2(), (Hull2d[jm]-Hull2d[gi]).norm2() );
			for ( gi=jp+1; gi!=jm; ++gi ) {
				if ( gi > n ) gi = 0; // loop around
				if ( max_sep.get_orig_ix(gi) != -1 ) { // this guy's still in
					break;
				}
			}
			//cout << "jp: " << jp << "; gi: " << gi << endl;
			max_sep[max_sep.get_orig_ix(jp)].first = std::max( (Hull2d[jm]-Hull2d[jp]).norm2(), (Hull2d[jp]-Hull2d[gi]).norm2() );

			/*cout << max_sep.size() << ":\n";
			for ( vector<pair<double,int> >::iterator i=max_sep.myvec.begin(); i!=max_sep.myvec.end(); ++i ) {
				cout << i->first << ": " << i->second << endl;
			}*/

			max_sep.pop_back(); // get rid of last element
		}
		/*cout << "DONE!\n" << max_sep.size() << ":\n";
		for ( vector<pair<double,int> >::iterator i=max_sep.myvec.begin(); i!=max_sep.myvec.end(); ++i ) {
			cout << i->first << ": " << i->second << endl;
		}*/
	}
}



