/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlRabbaniClustering.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
**
** DESCRIPTION:
**   Performs the Rabbani greedy region growing clustering technique.
*****************************************************************************/

#include "svlRabbaniClustering.h"

#define RAB_DEBUG_OUT 0

void svlRabbaniClustering::cluster(svlPointCloudData &PCD, vector<int> &asgs, vector<vector<int> > &R_ix, svlSortVectorDI &resid) {
	svlKdTree kdt(&PCD.pointCloud,20);
	int n_nn = 20;
	vector<vector<int> > knn(PCD.pointCloud.size(),vector<int>());
	for ( int i=0; i<(int)PCD.pointCloud.size(); ++i ) {
		kdt.findKNN(i,n_nn,-1.0,knn[i]);
	}
	cluster(PCD.pointCloud, PCD.pointNormals, knn, asgs, R_ix, resid);
}

void svlRabbaniClustering::cluster(svlPointCloudData &PCD, vector<vector<int> > &knn, vector<int> &asgs, vector<vector<int> > &R_ix, svlSortVectorDI &resid) {
	cluster(PCD.pointCloud, PCD.pointNormals, knn, asgs, R_ix, resid);
}

void svlRabbaniClustering::cluster(vector<svlPoint3d> &P, vector<svlPoint3d> &N, vector<vector<int> > &knn, vector<int> &asgs, vector<vector<int> > &R_ix, svlSortVectorDI &resid,
								   int mode, double r_th_in, double th_th_in, double th_th2_in) {
	// loop through lowest residuals
#if RAB_DEBUG_OUT
	FILE *fo; int foi; double fod;
#endif

	assert(P.size()==N.size());
	assert(P.size()==resid.size());
	assert(P.size()==resid._myvec.size());
	int n_pts = (int)P.size();
	double th_th, th_th2 = 0.83, r_th = 0.05;
	switch ( mode ) {
		case 0: case 2: th_th = 0.95; break;
		case 1: th_th = 0.7; break;
		case 3: th_th = 0.8; break;
		case 4: if ( r_th_in >= 0.0 )
					r_th = r_th_in;
				if ( th_th_in > 0.0 )
					th_th = th_th_in;
				if ( th_th2 > 0.0 )
					th_th2 = th_th2_in;
				else
					mode = 0; // do not use th2
				break;
		default: assert(false);
	}
	//printf("Mode: %d; r_th: %f; th_th: %f; th2_th: %f\n", mode, r_th, th_th, th_th2);
	bool *A = new bool[n_pts];
	for ( int i=0; i<n_pts; ++i ) A[i] = true;
	
	if ( !resid.is_sorted() )
	    resid.sort();

	// The algorithm.
	// Loop through the points through order of lowest residual.
	vector<pair<double,int> >::const_iterator ir = resid._myvec.begin();
	R_ix.clear();
	queue<int> S_c;
	int gi, since = 0, fi;
#define MIN_CLUS_SIZE 25 // need to see cluster of at least this size
#define CLUS_SIZE_HOLD 100 // within the last this many clusters

	bool old_mode = mode == 0 || mode == 1;
	while ( true ) {
		// find point in A with minimum residual
		while ( ir != resid._myvec.end() && !A[ir->second] ) ++ir;
		if ( ir == resid._myvec.end() )
			break; // finished with all points
		fi = ir->second;
		R_ix.push_back(vector<int>());
		R_ix.back().push_back(fi);
		S_c.push(fi);
		A[fi] = false;
		while ( !S_c.empty() ) { // loop through all seed points
			gi = S_c.front();
			S_c.pop();
			//int itr = 0;
			for ( vector<int>::const_iterator k=knn[gi].begin(); k!=knn[gi].end(); ++k ) {
				// loop over each seed point's nearest neighbors
				if ( A[*k] && fabs(N[gi].dot(N[*k])) >= th_th
						   && ( old_mode || fabs(N[fi].dot(N[*k])) >= th_th2 ) ) {
					R_ix.back().push_back(*k); // add point to region
					A[*k] = false; // mark point as added
					if ( resid.get_orig_val(*k) < r_th ) {
						S_c.push(*k);
						//S_c.push_back(*k); // add as a seed point to this group
					}
				}
			}
		}
		if ( (int)R_ix.back().size() < MIN_CLUS_SIZE ) // remove cluster
			R_ix.pop_back();
		++ir;
	}

	asgs = vector<int>(asgs.size(),-1); // init to -1's
	populateIx(R_ix,asgs);

	delete[] A;
}

void svlRabbaniClustering::cluster(svlPointCloudData &PCD, CvMat *Ix, vector<int> &asgs, vector<vector<int> > &R_ix, svlSortVectorDI &resid, int mode) {
	// loop through lowest residuals
	assert(PCD.pointCloud.size()==PCD.pointNormals.size());
	int n_pts = (int)PCD.pointCloud.size();
	double th_th, th_th2 = 0.83, r_th = 0.01;
	switch ( mode ) {
		case 0:
		case 2: th_th = 0.95; break;
		case 1: th_th = 0.7; break;
	}

#if RAB_DEBUG_OUT
	FILE *fo; int foi; double fod;
	char cha[32];
	sprintf(cha,"yyy_rab_%d.out",mode);
	fo = fopen(cha,"wb"); fclose(fo);
#endif

	vector<bool> A(n_pts,true); // checkable points
	if ( !resid.is_sorted() ) resid.sort();

	// The algorithm
	// loop through the points through order of lowest residual
	vector<pair<double,int> >::const_iterator ir = resid._myvec.begin();
	//vector<int> R_c, S_c; // new region and seeds
	R_ix.clear();
	queue<int> S_c;
	int gi, gi1, r, c, fi;
	if ( mode == 0 || mode == 1 ) {
		while ( true ) {
			// find point in A with minimum residual
			while ( ir != resid._myvec.end() && ir->second >= 0 && ir->second < n_pts && !A[ir->second] )
				++ir;
			if ( ir == resid._myvec.end() )
				break; // finished with all points
			R_ix.push_back(vector<int>());
			R_ix.back().push_back(ir->second);
			S_c.push(ir->second);
			fi = ir->second;
			A[ir->second] = false;
			while ( !S_c.empty() ) { // loop through all seed points
				gi = S_c.front();
				S_c.pop();
				r = gi/Ix->cols; // get r and c
				c = gi%Ix->cols;
				// check 5x5 neighborhood
	#define NHOOD 2
				for ( int r0=max(0,r-NHOOD); r0<=min(Ix->rows-1,r+NHOOD); ++r0 ) {
					for ( int c0=max(0,c-NHOOD); c0<=min(Ix->cols-1,c+NHOOD); ++c0 ) {
						if ( A[(gi1=c0+r0*Ix->cols)] && fabs(PCD.pointNormals[(mode==1?fi:gi)].dot(PCD.pointNormals[gi1])) >= th_th ) {
							R_ix.back().push_back(gi1); // add point to region
							A[gi1] = false; // mark point as added
							if ( resid.get_orig_val(gi1) < r_th )
								S_c.push(gi1);
						}
					}
				}
			}
			++ir;
		}
	} else {
		while ( true ) {
			// find point in A with minimum residual
			while ( ir != resid._myvec.end() && ir->second >= 0 && ir->second < n_pts && !A[ir->second] )
				++ir;
			if ( ir == resid._myvec.end() )
				break; // finished with all points
			R_ix.push_back(vector<int>());
			R_ix.back().push_back(ir->second);
			S_c.push(ir->second);
			fi = ir->second;
			A[ir->second] = false;
			while ( !S_c.empty() ) { // loop through all seed points
				gi = S_c.front();
				S_c.pop();
				r = gi/Ix->cols; // get r and c
				c = gi%Ix->cols;
				// check 5x5 neighborhood
	#define NHOOD 2
				for ( int r0=max(0,r-NHOOD); r0<=min(Ix->rows-1,r+NHOOD); ++r0 ) {
					for ( int c0=max(0,c-NHOOD); c0<=min(Ix->cols-1,c+NHOOD); ++c0 ) {
						if ( A[(gi1=c0+r0*Ix->cols)] && fabs(PCD.pointNormals[gi].dot(PCD.pointNormals[gi1])) >= th_th
													 && fabs(PCD.pointNormals[fi].dot(PCD.pointNormals[gi1])) >= th_th2 ) {
							R_ix.back().push_back(gi1); // add point to region
							A[gi1] = false; // mark point as added
							if ( resid.get_orig_val(gi1) < r_th )
								S_c.push(gi1);
						}
					}
				}
			}
			++ir;
		}
	}

	int j=0;
	for ( vector<vector<int> >::const_iterator j1=R_ix.begin(); j1!=R_ix.end(); ++j1 ) {
		for ( vector<int>::const_iterator k1=j1->begin(); k1!=j1->end(); ++k1 ) {
			if ( *k1 >= 0 && *k1 < (int)asgs.size() ) {
				asgs[*k1] = j;
			}
		}
		++j;
	}

	// ~
	A.clear();
}

