/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlGaussian.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  Implements a multi-variate gaussian distribution.
**
*****************************************************************************/

#pragma once

#include <vector>
#include <limits>

#include <cxcore.h>

#include "svlBase.h"

using namespace std;

// svlGaussian class ---------------------------------------------------------

class svlConditionalGaussian;

class svlGaussian {
 private:
    int _n;
    CvMat *_mu;               // n element mean vector
    CvMat *_Sigma;            // n-by-n element covariance matrix

    CvMat *_invSigma;         // n-by-n inverse covriance matrix
    float _logZ;              // log partition function (\sqrt(2*pi*det(Sigma)^n))

 public:
    svlGaussian(int n = 1);
    svlGaussian(const CvMat *mu, float sigma2);
    svlGaussian(const CvMat *mu, const CvMat *sigma2);
    svlGaussian(const vector<double>& mu, float sigma2);
    svlGaussian(const svlGaussian& model);
    ~svlGaussian();

    // initialization
    void initialize(int n);
    void initialize(const CvMat *mu, float sigma2);
    void initialize(const CvMat *mu, const CvMat *sigma2);
    void initialize(const vector<double>& mu, float sigma2);

    // conditioning
    svlGaussian reduce(const vector<double>& x,
        const vector<int>& indx) const;
    svlGaussian reduce(const map<int, double>& x) const;
    svlConditionalGaussian conditionOn(const vector<int>& indx) const;

    // i/o
    bool save(const char *filename) const;
    bool load(const char *filename);

    // evaluate (log-likelihood)
    void evaluate(const CvMat *x, CvMat *p) const;
    void evaluate(const vector<vector<double> >& x, vector<double>& p) const;
    double evaluateSingle(const CvMat *x) const;
    double evaluateSingle(const vector<double>& x) const;
    double evaluateSingle(double x) const;

    // learn parameters
    void train(const CvMat *x, float lambda = 0.0);
    void train(const vector<vector<double> >& x, float lambda = 0.0);

    // access
    const CvMat *mean() const { return _mu; }
    const CvMat *covariance() const { return _Sigma; }
    
    // standard operators
    svlGaussian& operator=(const svlGaussian& model);

 protected:
    void freeStorage();
    void updateCachedParameters();
};

// svlConditionalGaussian class ----------------------------------------------

class svlConditionalGaussian {
 private:
    int _n, _m;
    CvMat *_mu;               // n element mean vector \mu_1 - \Sigma_{12} \Sigma_{22}^{-1} \mu_2
    CvMat *_Sigma;            // n-by-n element covariance matrix \Sigma_{11} - \Sigma_{12} \Sigma_{22}^{-1} \Sigma_{21}
    CvMat *_SigmaGain;        // m-by-n gain matrix: \Sigma_{12} \Sigma_{22}^{-1}

    CvMat *_tmp;

 public:
    svlConditionalGaussian(const CvMat *mu, const CvMat *Sigma, 
        const CvMat *SigmaGain);
    svlConditionalGaussian(const svlConditionalGaussian& model);
    ~svlConditionalGaussian();

    // i/o
    bool save(const char *filename) const;
    bool load(const char *filename);

    // construct gaussian given observation
    svlGaussian reduce(const CvMat *x);
    svlGaussian reduce(const vector<double>& x);
};



