/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlObjectDetectionAnalyzer.h
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
**              based on applications by Stephen Gould <sgould@stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <sys/types.h>
#include <limits>

#include "svlBase.h"
#include "svlVision.h"

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#define _CRT_SECURE_NO_DEPRECATE
#include "win32/dirent.h"
#else
#include <dirent.h>
#endif

#define VERBOSE
#undef DEBUG
//#define DEBUG

using namespace std;


//------------------------------------------------------------------


static void filterDetections(  svlObject2dSequence &objects,
			       bool filterOverlapping = false,
			       bool filterByName = false,
			       string name = NO_NAME)
{
 if (filterByName && name != NO_NAME) {
    for (svlObject2dSequence::iterator it = objects.begin() ;
         it != objects.end(); ++it)
      removeNonMatchingObjects(it->second, name.c_str());
  }

  SVL_LOG(SVL_LOG_VERBOSE, objects.size() << " frames read");

  if (filterOverlapping) {
    int total = countObjects(objects);
    int removed = nonMaximalSuppression(objects, 0.25, 0.25, 0.64);
    SVL_LOG(SVL_LOG_VERBOSE, removed << " out of " << total << " non-maximal objects suppressed");
  }
}

static bool loadDetectionsFile(const char *filename,
			       svlObject2dSequence &objects,
			       bool filterOverlapping = false,
			       bool filterByName = false,
			       string name = NO_NAME) {

  if (!readObject2dSequence(filename, objects)) {
    SVL_LOG(SVL_LOG_ERROR, "...failed to read " << filename);
    return false;
  }

  filterDetections(objects, filterOverlapping, filterByName, name);
 
  return true;
}


bool svlObjectDetectionAnalyzer::loadGt(const char *filename) {
  return loadDetectionsFile(filename, _gtObjectList, false, true, _objectName);
}

bool svlMultObjectsDetectionAnalyzer::loadGt(const char *filename) {
  return loadDetectionsFile(filename, _gtObjectList);
}

bool svlObjectDetectionAnalyzer::loadDetections(const char *filename) {
  return loadDetectionsFile(filename, _objectList, _useNonmax, true, _objectName);
}

void svlObjectDetectionAnalyzer::loadDetections(const svlObject2dSequence & s)
{
  _objectList = s;
  filterDetections(_objectList,  _useNonmax, true, _objectName);
}

bool svlMultObjectsDetectionAnalyzer::loadDetections(const char *filename) {
  return loadDetectionsFile(filename, _objectList, _useNonmax);
}




//------------------------------------------------------------------------

void svlObjectDetectionAnalyzer::outputDetections(const char *filename,
						  float threshold) {
  svlObject2dSequence outputObjects;

  for (svlObject2dSequence::const_iterator it = _objectList.begin();
       it != _objectList.end(); ++it) {
    for (unsigned k = 0; k < it->second.size(); k++) {
      if (it->second[k].pr >= threshold) {
	outputObjects[it->first].push_back(it->second[k]);
      }
    }
  }

  writeObject2dSequence(filename, outputObjects);
}








//-----------------------------------------------------------------

static void updateAccumulatedScores(const svlObject2dFrame &detectedObjects,
				    const svlObject2dFrame &gtObjects,
				    float areaRatio,
				    svlClassifierResponseList &objectScores) {
  // list of objects already matched
  vector<bool> alreadyMatched(gtObjects.size(), false);

  // sort detections by score
  vector<pair<float, int> > sortedDetections;
  for (unsigned i = 0; i < detectedObjects.size(); i++) {
    sortedDetections.push_back(make_pair(-detectedObjects[i].pr, i));
  }
  stable_sort(sortedDetections.begin(), sortedDetections.end());
        
  // find overlapping detections
  for (unsigned i = 0; i < sortedDetections.size(); i++) {
    bool bMatched = false;
    svlObject2d detection = detectedObjects[sortedDetections[i].second];
    for (unsigned j = 0; j < alreadyMatched.size(); j++) {
      if (alreadyMatched[j])
	continue;
      if (gtObjects[j].areaOverlap(detection) < areaRatio)
	continue;
      
      // we have a match
      bMatched = true;
      alreadyMatched[j] = true;
      break;
    }
            
    insertScore(objectScores, detection.pr, bMatched);
  }

  // all detections not matched by now are misses
  for (unsigned i = 0; i < alreadyMatched.size(); i++) {
    if (!alreadyMatched[i]) {
      insertScore(objectScores, 0.0f, true);
    }
  }
}

// will not actually store the detections, but will continuously
// accumulate objectScores
bool svlObjectDetectionAnalyzer::accumulateScoresForFrame(string frameName,
						    svlObject2dFrame &detectedObjects,
						    float areaRatio) {
  svlObject2dSequence::const_iterator it = _gtObjectList.find(frameName);
  if (it == _gtObjectList.end())
    return false;

  const svlObject2dFrame &gtObjects = it->second;

  // filter the detections as you would when loading them from file
  int total = detectedObjects.size();
  int removed = nonMaximalSuppression(detectedObjects, 0.25, 0.25, 0.64);
  SVL_LOG(SVL_LOG_VERBOSE, removed << " out of " << total << " non-maximal objects suppressed");

  //cerr << removed << " out of " << total << " non-maximal objects suppressed" << endl;

  //for (int i = 0; i < detectedObjects.size(); i++)
  //cerr << detectedObjects[i].x << " " << detectedObjects[i].y << " " << detectedObjects[i].w << " " << detectedObjects[i].h << " " << detectedObjects[i].pr << endl;

  updateAccumulatedScores(detectedObjects, gtObjects, areaRatio, _objectScores);

  return true;
}






//-----------------------------------------------------------------

void svlObjectDetectionAnalyzer::accumulateScores(bool ignoreMissingFrames, float areaRatio)
{
    for (svlObject2dSequence::const_iterator itg = _gtObjectList.begin();
         itg != _gtObjectList.end(); ++itg) {
        
        const svlObject2dFrame& groundTruthObjects = itg->second;

        svlObject2dSequence::const_iterator itd = _objectList.find(itg->first);
        if (itd == _objectList.end()) {
            if (!ignoreMissingFrames) {
                // notch up missing for all objects in the frame
                for (unsigned i = 0; i < groundTruthObjects.size(); i++) {
		  if (groundTruthObjects[i].name == _objectName)
                    insertScore(_objectScores, 0.0f, true);
                }
            }
            continue;
        }

        const svlObject2dFrame detectedObjects = itd->second;

        // list of objects already matched
        vector<bool> alreadyMatched(groundTruthObjects.size(), false);

	updateAccumulatedScores(detectedObjects, groundTruthObjects, areaRatio, _objectScores);
    }
}


void svlMultObjectsDetectionAnalyzer::accumulateScores(bool ignoreMissingFrames,
						      float areaRatio)
{
  for (svlObject2dSequence::const_iterator itg = _gtObjectList.begin();
         itg != _gtObjectList.end(); ++itg) {
        
        const svlObject2dFrame& groundTruthObjects = itg->second;

        svlObject2dSequence::const_iterator itd = _objectList.find(itg->first);
        if (itd == _objectList.end()) {
            if (!ignoreMissingFrames) {
                // notch up missing for all objects in the frame
                for (unsigned i = 0; i < groundTruthObjects.size(); i++) {
		  insertScore(_allObjectScores[groundTruthObjects[i].name], 0.0f, true);
                }
            }
            continue;
        }

        const svlObject2dFrame& detectedObjects = itd->second;

        // list of objects already matched
        vector<bool> alreadyMatched(groundTruthObjects.size(), false);

        // sort detections by score
        vector<pair<float, int> > sortedDetections;
        for (unsigned i = 0; i < detectedObjects.size(); i++) {
            sortedDetections.push_back(make_pair(-detectedObjects[i].pr, i));
        }
        stable_sort(sortedDetections.begin(), sortedDetections.end());
        
        // find overlapping detections
        for (unsigned i = 0; i < sortedDetections.size(); i++) {
            bool bMatched = false;
            svlObject2d detection = detectedObjects[sortedDetections[i].second];

            for (unsigned j = 0; j < alreadyMatched.size(); j++) {
	      if (alreadyMatched[j])
                    continue;
                if (groundTruthObjects[j].name != detection.name)
                    continue;
                if (groundTruthObjects[j].areaOverlap(detection) < areaRatio)
                    continue;
                
                // we have a match
                bMatched = true;
                alreadyMatched[j] = true;
                break;
            }
            
            insertScore(_allObjectScores[detection.name], detection.pr, bMatched);
        }

        // all detections not matched by now are misses
        for (unsigned i = 0; i < alreadyMatched.size(); i++) {
            if (!alreadyMatched[i]) {
	      insertScore(_allObjectScores[itg->second[i].name], 0.0f, true);
            }
        }
    }
}



//---------------------------------------------------------

void svlObjectDetectionAnalyzer::printResults(float threshold) const  {
  cout << endl << "OBJECT\tCOUNT\tHIT\tERROR\tRECALL\tPREC.\tF1("
       << threshold << ")" << endl;
  cout << _objectName << "\t" << summary(_objectScores, threshold) << endl;
}

void svlMultObjectsDetectionAnalyzer::printResults(float threshold) const {
 cout << endl << "OBJECT\tCOUNT\tHIT\tERROR\tRECALL\tPREC.\tF1("
      << threshold << ")" << endl;
 for (TResponses::const_iterator it = _allObjectScores.begin();
      it != _allObjectScores.end(); ++it)
   cout << it->first << "\t" << summary(it->second, threshold) << endl;
}


void svlObjectDetectionAnalyzer::createPRcurves() {
  createPRcurve(_objectScores, _PRcurve);
}

void svlMultObjectsDetectionAnalyzer::createPRcurves() {
  for (TResponses::const_iterator it = _allObjectScores.begin();
       it != _allObjectScores.end(); ++it) {
    createPRcurve(it->second, _allPRcurves[it->first]);
  }
}


void svlObjectDetectionAnalyzer::createDiscretizedPRcurves(float dthreshold) {
  createDiscretizedPRcurve(_objectScores, _PRcurve, dthreshold);
}

void svlMultObjectsDetectionAnalyzer::createDiscretizedPRcurves(float dthreshold) {
  for (TResponses::const_iterator it = _allObjectScores.begin();
       it != _allObjectScores.end(); ++it)
    createDiscretizedPRcurve(it->second, _allPRcurves[it->first], dthreshold);
}


void svlObjectDetectionAnalyzer::writePRcurves(const char *filestem) const {
  string filename = string(filestem) + string(".") + _objectName + string(".txt");
  writePRcurve(filename.c_str(), _PRcurve);
}

void svlMultObjectsDetectionAnalyzer::writePRcurves(const char *filestem) const {
  for (TCurves::const_iterator it = _allPRcurves.begin();
       it != _allPRcurves.end(); ++it) {
    string filename = string(filestem) + string(".") + it->first + string(".txt");
    writePRcurve(filename.c_str(), it->second);
  }
}


bool svlObjectDetectionAnalyzer::loadPRcurve(const char *filename) {
  return readPRcurve(filename, _PRcurve);
}

bool svlMultObjectsDetectionAnalyzer::loadPRcurve(const char *filename) {
  SVL_LOG(SVL_LOG_WARNING, "Should provide a name for the PR curve");
  return false;
}

bool svlMultObjectsDetectionAnalyzer::loadPRcurve(const char *filename, string name) {
  if (name == NO_NAME)
    SVL_LOG(SVL_LOG_WARNING, "Should provide a non-empty name for the PR curve");
  return readPRcurve(filename, _allPRcurves[name]);
}

float svlObjectDetectionAnalyzer::computeHighestFscore() const {
  return getHighestFscore(_PRcurve);
}

float svlMultObjectsDetectionAnalyzer::computeHighestFscore(string name) {
  // no name is specified, try to guess
  if (name == NO_NAME) {
    if (_allPRcurves.size() == 1) {
      TCurves::iterator single_curve = _allPRcurves.begin();
      return getHighestFscore(single_curve->second);
    } else {
      SVL_LOG(SVL_LOG_WARNING, "No object specified, ambiguous");
      return 0;
    }
  }    
   
  // a name is given
  TCurves::iterator curve = _allPRcurves.find(name);
  if (curve != _allPRcurves.end()) {
    return getHighestFscore(curve->second);
  } else {
    SVL_LOG(SVL_LOG_WARNING, "No such object exists");
    return 0;
  }
}

float svlObjectDetectionAnalyzer::computePRarea() const {
  return getPRarea(_PRcurve);
}

float svlMultObjectsDetectionAnalyzer::computePRarea(string name) {
  // no name is specified, try to guess
  if (name == NO_NAME) {
    if (_allPRcurves.size() == 1) {
      TCurves::iterator curve = _allPRcurves.begin();
      return getPRarea(curve->second);
    } else {
      SVL_LOG(SVL_LOG_WARNING, "No object specified, ambiguous");
      return 0;
    }
  }
    
  // a name is given
  TCurves::iterator curve = _allPRcurves.find(name);
  if (curve != _allPRcurves.end()) {
    return getPRarea(curve->second);
  } else {
    SVL_LOG(SVL_LOG_WARNING, "No such object exists");
    return 0;
  }
}

void svlObjectDetectionAnalyzer::beginAccumulatingScores() {
  _objectScores.clear();
}

void svlMultObjectsDetectionAnalyzer::beginAccumulatingScores() { 
  _allObjectScores.clear();
}
















