/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlSlidingWindowDetector.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Sid Batra <sidbatra@stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <sys/types.h>
#include <limits>

#include "cxcore.h"
#include "ml.h"

#include "svlBase.h"
#include "svlVision.h"

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#define _CRT_SECURE_NO_DEPRECATE
#include "win32/dirent.h"
#else
#include <dirent.h>
#endif

#define VERBOSE
#undef DEBUG
//#define DEBUG

using namespace std;

// Statics and Globals -------------------------------------------------------

#ifndef DEBUG
bool svlSlidingWindowDetector::DEBUG_OUTPUT = false;
#else
bool svlSlidingWindowDetector::DEBUG_OUTPUT = true;
#endif

int svlSlidingWindowDetector::DELTA_X = 4;
int svlSlidingWindowDetector::DELTA_Y = 4;
double svlSlidingWindowDetector::DELTA_SCALE = 1.2;

static inline bool isSampleValid(vector<double> &sample) {
  for (unsigned i = 0; i < sample.size(); i++)
    if (isinf(sample[i]) || isnan(sample[i])) return false;
  return true;
}

// Constructors/Destructors --------------------------------------------------

svlSlidingWindowDetector::svlSlidingWindowDetector(const char *name, int w, int h) :
    svlOptions(), _model(NULL), 
    _windowWidth(w), _windowHeight(h), _threshold(0.5),
    _featureVector(NULL), _weakResponses(NULL)
{
    if (name == NULL) {
	_objectName = "[unknown]";
    } else {
	_objectName = string(name);
    }

    // default training options
    declareOption("boostingRounds", "100");
    declareOption("trimRate", "0.9");
    declareOption("numSplits", "2");
    declareOption("maxTrainingImages", numeric_limits<int>::max());

    // set the default deltas
    resetDeltas();
}

svlSlidingWindowDetector::svlSlidingWindowDetector(const svlSlidingWindowDetector& c) :
    svlOptions(c), _model(c._model),
    _windowWidth(c._windowWidth), _windowHeight(c._windowHeight), _objectName(c._objectName),
    _threshold(c._threshold), _featureVector(NULL), _weakResponses(NULL),
    _delta_x(c._delta_x), _delta_y(c._delta_y), _delta_scale(c._delta_scale)
{
    // do nothing
}

svlSlidingWindowDetector::~svlSlidingWindowDetector()
{
    if (_featureVector != NULL) {
        cvReleaseMat(&_featureVector);
    }

    if (_weakResponses != NULL) {
        cvReleaseMat(&_weakResponses);
    }
}

// Public Member Functions --------------------------------------------------

bool svlSlidingWindowDetector::readModel(const char *filename)
{
    assert(filename != NULL);

    if (_featureVector != NULL) {
        cvReleaseMat(&_featureVector);
        _featureVector = NULL;
    }
    if (_weakResponses != NULL) {
        cvReleaseMat(&_weakResponses);
        _weakResponses = NULL;
    }

    _model = svlSmartPointer<CvBoost>(new CvBoost());
    _model->load(filename);

    return (_model != NULL);
}

bool svlSlidingWindowDetector::writeModel(const char *filename)
{
    assert((_model != NULL) && (filename != NULL));
    ((CvBoost *)_model)->save(filename);
    return true;
}

// Classifies a multi-channel image.
svlObject2dFrame svlSlidingWindowDetector::classifyImage(const vector<IplImage *>& images)
{
  assert(_model != NULL);
  assert(!images.empty());
  
  // each image in the vector must not be NULL and all images must be the same size
  for (unsigned i = 0; i < images.size(); i++) 
    {
      assert(images[i] != NULL);
      assert(images[i]->width == images[0]->width);
      assert(images[i]->height == images[0]->height);
    }
  
  svlObject2dFrame objects;
  double scale = 1.0;
  float scoreThreshold = probabilityToScore((float)_threshold);
  int numChannels = (int)images.size();
  
  SVL_LOG(SVL_LOG_MESSAGE, "Running classifier on " << numChannels << "-channel image");
  
  // copy of original images for color conversion and rescaling
  //vector <IplImage *> originalImages(numChannels);
  vector <IplImage *> scaledImages(numChannels);
  
  for (int i = 0; i < numChannels; i++) 
    {
      scaledImages[i] = cvCloneImage(images[i]);
    }
  
  // iterate over all scales
  while ((scaledImages[0]->width >= _windowWidth) && (scaledImages[0]->height >= _windowHeight)) 
    {
      
      // iterate over all locations
      vector<CvPoint> locations;
      createWindowLocations(scaledImages[0]->width, scaledImages[0]->height, locations);
      
#if 1
      removeUninterestingLocations(scaledImages[0], locations);
#endif
	
	// run classifier
        SVL_LOG(SVL_LOG_VERBOSE, "Computing features over " << locations.size() << " windows for image size "
            << scaledImages[0]->width << " x " << scaledImages[0]->height << "...");

	Tsamples features;
	features.reserve(locations.size());
	computeImageFeatureVectors(features, locations, scaledImages);
		
	if (DEBUG_OUTPUT) {
	    string filename = string("svlSlidingWindowDetector.classifyImage.")
		+ toString(scale) + string(".txt");
	    ofstream ofs(filename.c_str(), ios::app);
	    for (unsigned i = 0; i < features.size(); i++) {
		ofs << (scale * locations[i].x) << " " << (scale * locations[i].y) << " "
		    << (scale * _windowWidth) << " " << (scale * _windowHeight);
		for (unsigned j = 0; j < features[i].size(); j++) {
		    ofs << " " << features[i][j];
		}
		ofs << "\n";
	    }
	    ofs.close();
	}

        SVL_LOG(SVL_LOG_VERBOSE, "...done");
        SVL_LOG(SVL_LOG_VERBOSE, "Running multi-channel classifier on " << locations.size() << " windows...");

	for (unsigned i = 0; i < locations.size(); i++) {
	  if(!isSampleValid(features[i])) continue;
	    float score = evaluateModel(features[i]);
	    if (score > scoreThreshold) {
		objects.push_back(svlObject2d(scale * locations[i].x, scale * locations[i].y,
                        scale * _windowWidth, scale * _windowHeight));
		objects.back().name = _objectName;
		objects.back().pr = scoreToProbability(score);
	    }
	}

	SVL_LOG(SVL_LOG_VERBOSE, "...done");

        // scale down each image channel
	scale *= _delta_scale;
	
        for (int i = 0; i < numChannels; i++) {
            cvReleaseImage(&(scaledImages[i]));
            //scaledImages[i] = cvCreateImage(cvSize((int)(originalImages[i]->width / scale),
	    //      (int)(originalImages[i]->height / scale)), originalImages[i]->depth, 1);
            //cvResize(originalImages[i], scaledImages[i]);
	    scaledImages[i] = cvCreateImage(cvSize((int)(images[i]->width / scale),
		    (int)(images[i]->height / scale)), images[i]->depth, images[i]->nChannels);
	    cvResize(images[i], scaledImages[i]);
        }
    }
    
    // free memory
    for (int i = 0; i < numChannels ; i++) {
        cvReleaseImage(&(scaledImages[i]));
        //cvReleaseImage(&(originalImages[i]));
    }
    
    return objects;
}



// Classifies a multi-channel image.
svlObject2dFrame svlSlidingWindowDetector::classifyImage(const vector<IplImage *>& images,
							 const map<double, vector<CvPoint> > &locationsWithScales,
							 bool sparse)
{
  assert(_model != NULL);
  assert(!images.empty());
  
  // each image in the vector must not be NULL and all images must be the same size
  for (unsigned i = 0; i < images.size(); i++) 
    {
      assert(images[i] != NULL);
      assert(images[i]->width == images[0]->width);
      assert(images[i]->height == images[0]->height);
    }

  int numChannels = (int)images.size();
  
  svlObject2dFrame objects;
  
  SVL_LOG(SVL_LOG_MESSAGE, "Running classifier on " << numChannels << "-channel image");
  
  vector <IplImage *> scaledImages(numChannels);
  for (int i = 0; i < numChannels; i++)
    scaledImages[i] = NULL;
  
  for (map<double, vector<CvPoint> >::const_iterator it = locationsWithScales.begin();
       it != locationsWithScales.end();
       it++) {

    double scale = it->first;

    for (int i = 0; i < numChannels; i++) {
      if (scaledImages[i])
	cvReleaseImage(&(scaledImages[i]));
      scaledImages[i] = cvCreateImage(cvSize((int)(images[i]->width / scale),
	    (int)(images[i]->height / scale)), images[i]->depth, images[i]->nChannels);
      cvResize(images[i], scaledImages[i]);
    }

    if ((scaledImages[0]->width < _windowWidth) || (scaledImages[0]->height < _windowHeight)) 
      continue;

    svlObject2dFrame new_objects = classifyImage(scaledImages, it->second, sparse);
    scaleObject2dFrame(new_objects, scale);
    
    objects.insert(objects.end(), new_objects.begin(), new_objects.end());
  }

  // free memory
  for (int i = 0; i < numChannels ; i++) {
    if (scaledImages[i])
      cvReleaseImage(&(scaledImages[i]));
  }

  return objects;
}



void svlSlidingWindowDetector::getAllWindows(int width, int height, svlObject2dFrame & windows) const
{

  int scaledWidth = width;
  int scaledHeight = height;

  double scale = 1.0;

  //iterate over all scales
  while ((scaledWidth >= _windowWidth) && (scaledHeight >= _windowHeight))
    {
      //iterate over all locations
      vector<CvPoint> locations;
      createWindowLocations(scaledWidth, scaledHeight, locations);

      for (unsigned i = 0; i < locations.size(); i++)
	windows.push_back( svlObject2d(locations[i].x * scale, locations[i].y * scale, _windowWidth * scale, _windowHeight * scale));

      scale *= this->DELTA_SCALE;

      scaledWidth = (int)(width / scale);
      scaledHeight = (int)(height / scale);
      
    }

}

void svlSlidingWindowDetector::removeInvalidLocations(IplImage *image, vector<CvPoint> &locations) {
  for (int i = (int)locations.size() - 1; i >= 0; i--) 
    {
      if (locations[i].x < 0 || locations[i].y < 0 ||
	  locations[i].x + _windowWidth > image->width || locations[i].y + _windowHeight > image->height) {
	locations.erase(locations.begin() + i);
      }
    }
}

void svlSlidingWindowDetector::removeUninterestingLocations(IplImage *image, vector<CvPoint> &locations) {
  for (int i = (int)locations.size() - 1; i >= 0; i--) 
    {
      cvSetImageROI(image, cvRect(locations[i].x, locations[i].y, _windowWidth, _windowHeight));
      if (isImageAlmostUniform(image))
	locations.erase(locations.begin() + i);
      cvResetImageROI(image);
    }
}

svlObject2dFrame svlSlidingWindowDetector::classifyImage(const vector<IplImage *>& images,
							 const vector<CvPoint>& locationsOrig,
							 bool sparse)
{
    assert((_model != NULL) && (!images.empty()));
    // each image in the vector must not be NULL and all images must be the same size
    for (unsigned i = 0; i < images.size(); i++) {
        assert(images[i] != NULL);
        assert(images[i]->width == images[0]->width);
        assert(images[i]->height == images[0]->height);
    }

    svlObject2dFrame objects;
    float scoreThreshold = probabilityToScore((float)_threshold);

    vector<CvPoint> locations = locationsOrig;

    removeInvalidLocations(images[0], locations);

#if 1
    // check variation within region
    removeUninterestingLocations(images[0], locations);
#endif

    // run classifier
    SVL_LOG(SVL_LOG_VERBOSE, "Computing features over " << locations.size() << " windows for image size "
        << images[0]->width << " x " << images[0]->height << "...");

    
    Tsamples features;
    features.reserve(locations.size());
	
    computeImageFeatureVectors(features, locations, images, sparse);
	
    SVL_LOG(SVL_LOG_VERBOSE, "...done");
    SVL_LOG(SVL_LOG_VERBOSE, "Running multi-channel classifier on " << locations.size() << " windows...");

    for (unsigned i = 0; i < locations.size(); i++) 
	{
	  if(!isSampleValid(features[i])) continue;
	  float score = evaluateModel(features[i]);
	  if (score > scoreThreshold) {
            objects.push_back(svlObject2d(locations[i].x, locations[i].y,
					  _windowWidth, _windowHeight));
            objects.back().name = _objectName;
            objects.back().pr = scoreToProbability(score);
	  }
	}
    
    SVL_LOG(SVL_LOG_VERBOSE, "...done");
    
    return objects;
}

// evaluate classifier on given region (and subwindows)
svlObject2dFrame svlSlidingWindowDetector::classifySubImage(const vector<IplImage *>& images,
    CvRect region)
{
    assert((_model != NULL) && (!images.empty()));
    // each image in the vector must not be NULL and all images must be the same size
    for (unsigned i = 0; i < images.size(); i++) {
        assert(images[i] != NULL);
        assert(images[i]->width == images[0]->width);
        assert(images[i]->height == images[0]->height);
    }

    // if entire image, run full detector
    if ((region.x == 0) && (region.width == images[0]->width) &&
	(region.y == 0) && (region.height == images[0]->height)) {
	return classifyImage(images);
    }

    int numChannels = (int)images.size();

    // create subimages
    vector <IplImage *> subImages(numChannels);
    svlClipRect(region, images[0]);
    for (int i = 0; i < numChannels; i++) {
        IplImage *imgCopy = cvCloneImage(images[i]);

        subImages[i] = cvCreateImage(cvSize(region.width, region.height),
            images[i]->depth, images[i]->nChannels);
        assert(subImages[i] != NULL);
        
        cvSetImageROI(imgCopy, region);
        cvCopyImage(imgCopy, subImages[i]);
        cvReleaseImage(&imgCopy);
    }
    
    // run full detector on sub-images
    svlObject2dFrame objects = classifyImage(subImages);

    for (unsigned i = 0; i < objects.size(); i++) {
	objects[i].x += region.x;
	objects[i].y += region.y;
    }

    // free memory
    for (int i = 0; i < numChannels; i++) {
        cvReleaseImage(&subImages[i]);
    }

    return objects;
}

// create sliding window locations for given scale ----------------------------

void svlSlidingWindowDetector::createWindowLocations(int width, int height,
    vector<CvPoint>& locations) const
{
    int maxX = width - _windowWidth;
    int maxY = height - _windowHeight;

    if ((maxX <= 0) || (maxY <= 0))
	return;

    locations.reserve((maxX + _delta_x) * (maxY + _delta_y) / (_delta_x * _delta_y));
    for (int x = 0; x <= maxX; x += _delta_x) {
	for (int y = 0; y <= maxY; y += _delta_y) {
	    locations.push_back(cvPoint(x, y));
	}
    }
}

void svlSlidingWindowDetector::createAllWindowLocations(int width, int height,
    map<double, vector<CvPoint> >& locations) const
{
    double scale = 1.0;
    int w = width;
    int h = height;
    locations.clear();
    while ((w >= _windowWidth) && (h >= _windowHeight)) {
	createWindowLocations(w, h, locations[scale]);
	scale *= _delta_scale;
	w = (int)(width / scale);
	h = (int)(height / scale);
    }    
}

// learning functions -------------------------------------------------------

static inline int makeSamplesValid(Tsamples& samples) {
  int orig_size = samples.size();
  for (int i = (int)samples.size() - 1; i >= 0; i--) {
    if (!isSampleValid(samples[i]))
      samples.erase(samples.begin() + i);
  }
  // returns the number of samples erased
  return orig_size - samples.size();
}

static void createTrainingSamples(Tsamples &orig, float num_in_batch, int fold,
				  Tsamples &newSamples,
				  Tsamples &newCVsamples) {
  newSamples.clear();
  newCVsamples.clear();

  Tsamples::iterator cutoff1 = orig.begin() + (int)(num_in_batch * fold);
  Tsamples::iterator cutoff2 = orig.begin() + (int)(num_in_batch * (fold+1));

  newSamples.insert(newSamples.end(), orig.begin(), cutoff1);
  newCVsamples.insert(newCVsamples.end(), cutoff1, cutoff2);
  newSamples.insert(newSamples.end(), cutoff2, orig.end());
  
}

static float computeFscore(int tp, int fp, int num_pos, int num_neg, bool useNormedF) {
  float precision = 0;
  float recall = 0;
  
  float Fscore = 0;
  
  if (num_pos && num_neg)
    {    
      if (useNormedF)
	{
	  float tp_norm = (float)tp / num_pos;
	  float fp_norm = (float)fp / num_neg;
	  
	  recall = (float)tp_norm;
	  if (tp_norm + fp_norm > 0) precision = (float)tp_norm / (tp_norm + fp_norm);
	  	  
	}
      else
	{
	  recall = ((float) tp) / (float) (num_pos);
	  if (tp+fp > 0) precision = ((float) tp) / (float) (tp+fp);
	}
      
      if (precision > 0 && recall > 0)
	Fscore =  2.0f / ((1.0f/precision) + (1.0f/recall));
    }

  return Fscore;
}

void svlSlidingWindowDetector::learnModelFullCV(Tsamples& posSamples,
						Tsamples& negSamples,
						int kfold) {
  if (kfold <= 1) {
    learnModel(posSamples, negSamples);
    return;
  }

#if 1
    // erase non-valid samples
    int num_erased = makeSamplesValid(posSamples);
    if (num_erased > 0) SVL_LOG(SVL_LOG_WARNING, "Had to erase " << num_erased << " invalid positive samples");
    num_erased = makeSamplesValid(negSamples);
    if (num_erased > 0) SVL_LOG(SVL_LOG_WARNING, "Had to erase " << num_erased << " invalid negative samples");
#endif

  int num_pos = posSamples.size();
  float num_pos_in_batch = (float)num_pos / kfold;

  int num_neg = negSamples.size();
  float num_neg_in_batch = (float)num_neg / kfold;

  Tsamples pos, neg, posCV, negCV;

  int num_rounds = getOptionAsInt("boostingRounds");
  int num_reeval_rounds = 20;
  int num_options = (int)ceil((double)num_rounds / (double)num_reeval_rounds);

  vector<int> tp(num_options, 0);
  vector<int> fp(num_options, 0);

  for (int i = 0; i < kfold; i++) {
    createTrainingSamples(posSamples, num_pos_in_batch, i, pos, posCV);
    createTrainingSamples(negSamples, num_neg_in_batch, i, neg, negCV);

    learnModel(pos, neg, false);

    for (int j = 0; j < num_options; j++) {
      int cur_rounds = min((j+1) * num_reeval_rounds, num_rounds);
      CvSlice slice = cvSlice(0, cur_rounds);

      // otherwise the evaluation code will not work, since the number of model
      // parameters will have changed
      if (_weakResponses) {
	cvReleaseMat(&_weakResponses);
	_weakResponses = NULL;
      }
      
      tp[j] += numClassifiedAsPositive(posCV, slice);
      fp[j] += numClassifiedAsPositive(negCV, slice);
    }
  }

  float best_option = -1;
  int best_i = -1;
  int iter = 0;
  int max_iters_without_improvement = 3;
  for (int i = 0; i < num_options; i++) {
    float option = computeFscore(tp[i], fp[i], posSamples.size(), negSamples.size(), true);
    SVL_LOG(SVL_LOG_MESSAGE, "After " << (i+1)*num_reeval_rounds << " rounds of training, "
	    << "performance (normed) is " << option);
    if (option > best_option) {
      best_option = option;
      best_i = i;
      iter = 0;
    } else {
      iter++;
    }
    if (iter >= max_iters_without_improvement)
      break;
  }

  setOption("boostingRounds", (best_i + 1) * num_reeval_rounds);

  SVL_LOG(SVL_LOG_MESSAGE, "Chose " << (best_i + 1) * num_reeval_rounds << " as the best number of boosting rounds");
  
  learnModel(posSamples, negSamples);
}


void svlSlidingWindowDetector::learnModel(Tsamples& posSamples,
					  Tsamples& negSamples,
					  bool print_results)
{
    if (_featureVector != NULL) {
        cvReleaseMat(&_featureVector);
        _featureVector = NULL;
    }
    if (_weakResponses != NULL) {
        cvReleaseMat(&_weakResponses);
        _weakResponses = NULL;
    }

#if 1
    // erase non-valid samples
    int num_erased = makeSamplesValid(posSamples);
    if (num_erased > 0) SVL_LOG(SVL_LOG_WARNING, "Had to erase " << num_erased << " invalid positive samples");
    num_erased = makeSamplesValid(negSamples);
    if (num_erased > 0) SVL_LOG(SVL_LOG_WARNING, "Had to erase " << num_erased << " invalid negative samples");
#endif


    SVL_LOG(SVL_LOG_MESSAGE, "Training with " << posSamples.size() << " positive and "
        << negSamples.size() << " negative samples");
    SVL_LOG(SVL_LOG_MESSAGE, "Feature vector size is " << posSamples[0].size());

#ifdef DEBUG
    ofstream ofs("svlObjectDetector.learnModel.txt");
    for (unsigned i = 0; i < posSamples.size(); i++) {
	for (unsigned j = 0; j < posSamples[i].size(); j++) {
	    ofs << posSamples[i][j] << " ";
	}
	ofs << 1 << endl;
    }
    for (unsigned i = 0; i < negSamples.size(); i++) {
	for (unsigned j = 0; j < negSamples[i].size(); j++) {
	    ofs << negSamples[i][j] << " ";
	}
	ofs << 0 << endl;
    }    
    ofs.close();
#endif

    // assemble design matrix for training
    CvMat *data = cvCreateMat((int)(posSamples.size() + negSamples.size()),
        (int)posSamples[0].size(), CV_32FC1);
    CvMat *labels = cvCreateMat((int)(posSamples.size() + negSamples.size()), 1, CV_32SC1);
    assert((data != NULL) && (labels != NULL));

    for (unsigned i = 0; i < posSamples.size(); i++) {
	SVL_ASSERT_MSG(posSamples[i].size() == (unsigned)data->width, "Failed on "
		       << i << "th sample (out of " << posSamples.size() << "), with "
		       << "data->width: " << data->width << " and posSamples[i].size(): "
		       << posSamples[i].size());

	for (unsigned j = 0; j < (unsigned)data->width; j++) {
	    cvmSet(data, i, j, posSamples[i][j]);
	}
	CV_MAT_ELEM(*labels, int, i, 0) =  1;
    }

    for (unsigned i = 0; i < negSamples.size(); i++) {
      SVL_ASSERT_MSG(negSamples[i].size() == (unsigned)data->width, "Failed on "
		     << i << "th sample (out of " << negSamples.size() << "), with "
		     << "data->width: " << data->width << " and negSamples[i].size(): "
		     << negSamples[i].size());

	for (unsigned j = 0; j < (unsigned)data->width; j++) {
	    cvmSet(data, i + posSamples.size(), j, negSamples[i][j]);
	}
	CV_MAT_ELEM(*labels, int, i + posSamples.size(), 0) = -1;
    }

    CvMat *varType = cvCreateMat(data->width + 1, 1, CV_8UC1);
    for (int i = 0; i < data->width; i++) {
      CV_MAT_ELEM(*varType, unsigned char, i, 0) = CV_VAR_NUMERICAL;
    }
    CV_MAT_ELEM(*varType, unsigned char, data->width, 0) = CV_VAR_CATEGORICAL;


    // train the classifier
    _model = svlSmartPointer<CvBoost>(new CvBoost());
    
    CvBoostParams parameters(CvBoost::GENTLE, getOptionAsInt("boostingRounds"), getOptionAsDouble("trimRate"), \
			     getOptionAsInt("numSplits"), false, NULL);

    SVL_LOG(SVL_LOG_VERBOSE, "Creating boosting model...");
    _model->train(data, CV_ROW_SAMPLE, labels, NULL, NULL, varType, NULL, parameters);
    SVL_LOG(SVL_LOG_VERBOSE, "...done");

    // otherwise the evaluation code will not work, since the number of model
    // parameters might have changed
    if (_weakResponses) {
      cvReleaseMat(&_weakResponses);
      _weakResponses = NULL;
    }

    // evaluate model (on training data) and print the results
    if (print_results)
      modelPerformance(posSamples, negSamples, "", true);

    cvReleaseMat(&varType);
    cvReleaseMat(&labels);
    cvReleaseMat(&data);    
}

void svlSlidingWindowDetector::learnModelCV(Tsamples& posSamples,
					    Tsamples& negSamples, 
					    Tsamples& posCVsamples,
					    Tsamples& negCVsamples,
					    bool retrain)
{
  learnModel(posSamples, negSamples, false);
 
#if 1
    int num_erased = makeSamplesValid(posCVsamples);
    if (num_erased > 0) SVL_LOG(SVL_LOG_WARNING, "Had to erase " << num_erased << " invalid positive cv samples");
    num_erased = makeSamplesValid(negCVsamples);
    if (num_erased > 0) SVL_LOG(SVL_LOG_WARNING, "Had to erase " << num_erased << " invalid negative cv samples");
#endif

  SVL_LOG(SVL_LOG_MESSAGE, "Using " << posCVsamples.size() << " positive and "
	  << negCVsamples.size() << " negative cross-validation samples");

  int num_rounds = getOptionAsInt("boostingRounds");
  
  float best_score = -1;
  int best_num_rounds = -1;
  
  while (num_rounds > 0) {    
    CvSlice slice = cvSlice(0, num_rounds);

    SVL_LOG(SVL_LOG_MESSAGE, "\tAfter " << num_rounds << " rounds of training");

    // otherwise the evaluation code will not work, since the number of model
    // parameters will have changed
    if (_weakResponses) {
      cvReleaseMat(&_weakResponses);
      _weakResponses = NULL;
    }
 
    // evaluate model (on training data) and print the results
    float training_performance = modelPerformance(posSamples, negSamples, "", true, slice);
      
    // evaluate model (on cv data)
    float score = modelPerformance(posCVsamples, negCVsamples, "on cv", true, slice);  

    // two things going on here:
    // (1) makes sure that a random spike in cv performance at the beginning of
    //     training doesn't screw it up (has happened before); thus wait either
    //     until 100 rounds have passed, or training performance gets to 1, which
    //     it almost inevitably will (this is hacky, but necessary for now)
    // (2) given the option between two equally-scoring rounds, want to choose
    //     the smaller one, so score >= best_score, not score > best_score
    if ((num_rounds >= 100 || training_performance > 0.99999) &&
	(score >= best_score)) {
      best_score = score;
      best_num_rounds = num_rounds;
    }
	  
    // with fewer weak learners, each one matters much more
    int num_reeval_rounds = 20;
    if (num_rounds <= 200)
      num_reeval_rounds = 10;
    if (num_rounds <= 100)
      num_reeval_rounds = 5;
    
    num_rounds -= num_reeval_rounds;
  }

  // release memory
  if (_weakResponses) {
    cvReleaseMat(&_weakResponses);
    _weakResponses = NULL;
  }

  //   re-train the model with either
  // entire train+cv set, or just the cv set
  //   this is important because the number of boosting rounds chosen
  // is different from the last trained model 
  setOption("boostingRounds", best_num_rounds);
      
  if (retrain) {
    SVL_LOG(SVL_LOG_MESSAGE, "\tRetraining");
    unsigned orig_num_pos = posSamples.size();
    unsigned orig_num_neg = negSamples.size();
    
    Tsamples fullPosSet(orig_num_pos + posCVsamples.size());
    Tsamples fullNegSet(orig_num_neg + negCVsamples.size());
    
    for (unsigned i = 0; i < orig_num_pos; i++)
      fullPosSet[i] = posSamples[i];
    for (unsigned i = 0; i < posCVsamples.size(); i++)
      fullPosSet[orig_num_pos + i] = posCVsamples[i];
    
    for (unsigned i = 0; i < orig_num_neg; i++)
      fullNegSet[i] = negSamples[i];
    for (unsigned i = 0; i < negCVsamples.size(); i++)
      fullNegSet[orig_num_neg + i] = negCVsamples[i];
    
    learnModel(fullPosSet, fullNegSet);
  }
  else {
    //learnModel(posSamples, negSamples);
    _model->prune(cvSlice(0, best_num_rounds));
  }
}

int svlSlidingWindowDetector::numClassifiedAsPositive(const Tsamples &samples,
						      CvSlice slice) {
  int result = 0;
  for (unsigned i = 0; i < samples.size(); i++) {
    if (evaluateModel(samples[i], slice) > 0.0)
      result++;
  }
  return result;
}

float svlSlidingWindowDetector::modelPerformance(const Tsamples& posSamples,
						 const Tsamples& negSamples,
						 const string label, bool useNormedF,
						 CvSlice slice) {
  int tp = 0;
  int tn = 0;
  int fp = 0;
  int fn = 0;
  for (unsigned i = 0; i < posSamples.size(); i++) {
    if (evaluateModel(posSamples[i], slice) > 0.0) {
      tp += 1;
    } else {
      fn += 1;
    }
  }
  for (unsigned i = 0; i < negSamples.size(); i++) {
    if (evaluateModel(negSamples[i], slice) > 0.0) {
      fp += 1;
    } else {
      tn += 1;
    }
  }

  float Fscore = computeFscore(tp, fp, posSamples.size(), negSamples.size(), useNormedF);
    
  if (useNormedF) {
    SVL_LOG(SVL_LOG_MESSAGE, "Evaluation " << label << " (TP, FN, TN, FP, normed \"F\" score): " << tp << ", " << fn << ", " << tn << ", " << fp << ", " << Fscore);
  } else {
    SVL_LOG(SVL_LOG_MESSAGE, "Evaluation " << label << " (TP, FN, TN, FP, F-score): " << tp << ", " << fn << ", " << tn << ", " << fp << ", " << Fscore);  
  }

  return Fscore;
}

float svlSlidingWindowDetector::evaluateModel(const vector<double>& fv, CvSlice slice)
{
  assert(_model != NULL);
  if (_featureVector == NULL) {
    _featureVector = cvCreateMat(1, (int)fv.size(), CV_32FC1);
  }
  if (_weakResponses == NULL) {
    int length = cvSliceLength(slice, _model->get_weak_predictors());
    _weakResponses = cvCreateMat(length, 1, CV_32FC1);
  }
  
  for (int j = 0; j < _featureVector->cols; j++) {
    cvmSet(_featureVector, 0, j, fv[j]);
  }
  
  
  _model->predict(_featureVector, NULL, _weakResponses, slice);
  float score = (float)cvSum(_weakResponses).val[0];
  
  return score;
}

float svlSlidingWindowDetector::evaluateModel(const vector<IplImage *>& images, CvRect region)
{
    vector<double> fv;

    assert((_model != NULL) && (!images.empty()));
    // each image in the vector must not be NULL and all images must be the same size
    for (unsigned i = 0; i < images.size(); i++) {
        assert(images[i] != NULL);
        assert(images[i]->width == images[0]->width);
        assert(images[i]->height == images[0]->height);
    }

    // if entire image, run full detector
    if ((region.x == 0) && (region.width == images[0]->width) &&
	(region.y == 0) && (region.height == images[0]->height)) {
        computePatchFeatureVector(fv, images);
        return evaluateModel(fv);
    }

    int numChannels = (int)images.size();

    // create subimages
    vector <IplImage *> subImages(numChannels);
    svlClipRect(region, images[0]);
    for (int i = 0; i < numChannels; i++) {
        IplImage *imgCopy = cvCloneImage(images[i]);

        subImages[i] = cvCreateImage(cvSize(region.width, region.height),
            images[i]->depth, images[i]->nChannels);
        assert(subImages[i] != NULL);
        
        cvSetImageROI(imgCopy, region);
        cvCopyImage(imgCopy, subImages[i]);
        cvReleaseImage(&imgCopy);
    }
    
    // compute feature vector
    computePatchFeatureVector(fv, subImages);

    // free memory
    for (int i = 0; i < numChannels; i++) {
        cvReleaseImage(&subImages[i]);
    }

    return evaluateModel(fv);
}

float svlSlidingWindowDetector::scoreToProbability(float score) const
{
  return 1.0f / (1.0f + exp(-score));
}

float svlSlidingWindowDetector::probabilityToScore(float pr) const
{
    return (float)(log(pr) - log(1.0f - pr));
}

// svlSingleChannelSlidingWindowDetector -------------------------------------

svlSingleChannelSlidingWindowDetector::svlSingleChannelSlidingWindowDetector() :
    svlSlidingWindowDetector()
{
    // do nothing
}


svlSingleChannelSlidingWindowDetector::~svlSingleChannelSlidingWindowDetector()
{
    // do nothing
}

// evaluate classifer over all shifts and scales
svlObject2dFrame svlSingleChannelSlidingWindowDetector::classifyImage(IplImage *image)
{
    vector<IplImage*> v(1, image);
    return svlSlidingWindowDetector::classifyImage(v);
}

// evaluate classifier at base scale and given shifts
svlObject2dFrame svlSingleChannelSlidingWindowDetector::classifyImage(IplImage *image,
			      const std::vector<CvPoint>& locations, bool sparse)
{
    vector<IplImage*> v(1, image);
    return svlSlidingWindowDetector::classifyImage(v, locations, sparse);
}

// evaluate classifier on given region (and subwindows)
svlObject2dFrame svlSingleChannelSlidingWindowDetector::classifySubImage(IplImage *image,
    CvRect region)
{
    vector<IplImage*> v(1, image);
    return svlSlidingWindowDetector::classifySubImage(v, region);
}

// evaluate the model on a given feature vector/patch
float svlSingleChannelSlidingWindowDetector::evaluateModel(IplImage * image,
    CvRect region)
{
    vector<IplImage*> v(1, image);
    return svlSlidingWindowDetector::evaluateModel(v, region);
}





































#if 0

void svlSlidingWindowDetector::learnModel(const char *posDirectory,  const char *negDirectory,
					  const char *extensions[], int numChannels)
{

  SVL_LOG(SVL_LOG_FATAL, "function not properly implemented currently");

    Tsamples posSamples;
    Tsamples negSamples;

    unsigned maxTrainingImages = (unsigned)getOptionAsInt("maxTrainingImages");

    // read, resize and process positive images 
    DIR *dir = opendir(posDirectory);    
    if (dir == NULL) {
        SVL_LOG(SVL_LOG_FATAL, "could not open image directory " << posDirectory);
    }

    // load images (based on first channel image existing)
    struct dirent *e = readdir(dir);
    while (e != NULL) {
        if (strstr(e->d_name, extensions[0]) != NULL) {
            SVL_LOG(SVL_LOG_MESSAGE, "Processing " << e->d_name << "...");
    		            
            // holds all channels for each image 
            vector<IplImage*> imgs;

            for (int i = 0; i < numChannels; i++) {
                // form filename of the image for the current channel
                string imageFilename = string(e->d_name);
                imageFilename.replace(imageFilename.length() - strlen(extensions[0]),
                    strlen(extensions[0]), string(extensions[i]));
                imageFilename = string(posDirectory) + string("/") + imageFilename;

                // TO DO: make this more generic
                IplImage *img = NULL;				
                if (i == 0) {
                    img = cvLoadImage(imageFilename.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
                } else {
                    img = readMatrixAsIplImage(imageFilename.c_str(), _windowWidth, _windowHeight);
                }    		
                assert(img != NULL);
                
                // resize image
                if ((img->width != _windowWidth) || (img->height != _windowHeight)) {
                    IplImage *resizedImage = NULL;
                    resizedImage = cvCreateImage(cvSize(_windowWidth, _windowHeight), img->depth, 1);
                    
                    cvResize(img, resizedImage);
                    cvReleaseImage(&img);
                    img = resizedImage;
                }

                // store for feature vector computation
                imgs.push_back(img);
            } // loop over numChannels
                     
            // process image
            posSamples.push_back(vector<double>());
            computePatchFeatureVector(posSamples.back(), imgs);
				
            // free memory
            for (int i = 0; i < numChannels; i++) {
                cvReleaseImage(&(imgs[i]));
            }
        }

        if (posSamples.size() >= maxTrainingImages) {
            break;
        }

        e = readdir(dir);	
    } // next image
    closedir(dir);
    
    // read, resize and process negative images 
    dir = opendir(negDirectory);
    if (dir == NULL) {
        SVL_LOG(SVL_LOG_FATAL, "could not open image directory " << negDirectory);
    }

    e = readdir(dir);
    while (e != NULL) {
        if (strstr(e->d_name, extensions[0]) != NULL) {
            SVL_LOG(SVL_LOG_MESSAGE, "Processing " << e->d_name << "...");
    		            
            // holds all channels for each image 
            vector<IplImage*> imgs;

            for (int i = 0; i < numChannels; i++) {
                // form filename of the image for the current channel
                string imageFilename = string(e->d_name);
                imageFilename.replace(imageFilename.length() - strlen(extensions[0]),
                    strlen(extensions[0]), string(extensions[i]));
                imageFilename = string(posDirectory) + string("/") + imageFilename;

                // TO DO: make this more generic
                IplImage *img = NULL;				
                if (i == 0) {
                    img = cvLoadImage(imageFilename.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
                } else {
                    img = readMatrixAsIplImage(imageFilename.c_str(), _windowWidth, _windowHeight);
                }    		
                assert(img != NULL);
                
                // resize image
                if ((img->width != _windowWidth) || (img->height != _windowHeight)) {
                    IplImage *resizedImage = NULL;
                    resizedImage = cvCreateImage(cvSize(_windowWidth, _windowHeight), img->depth, 1);
                    
                    cvResize(img, resizedImage);
                    cvReleaseImage(&img);
                    img = resizedImage;
                }

                // store for feature vector computation
                imgs.push_back(img);
            } // loop over numChannels
                     
            // process image
            negSamples.push_back(vector<double>());
            computePatchFeatureVector(negSamples.back(), imgs);
				
            // free memory
            for (int i = 0; i < numChannels; i++) {
                cvReleaseImage(&(imgs[i]));
            }
        }

        if (negSamples.size() >= maxTrainingImages) {
            break;
        }

        e = readdir(dir);	
    } // next image
    closedir(dir);

    // learn the model
    learnModel(posSamples, negSamples);
}


void svlSlidingWindowDetector::learnModel(Tsamples& posSamples,
					  Tsamples& negSamples)
{
  Tsamples tmp_pos;
  Tsamples tmp_neg;

  learnModel(posSamples, negSamples, tmp_pos, tmp_neg, false);
}

void svlSlidingWindowDetector::learnModel(Tsamples& posSamples,
					  Tsamples& negSamples, 
					  Tsamples& posCVsamples,
					  Tsamples& negCVsamples,
					  bool retrain)
{
    if (_featureVector != NULL) {
        cvReleaseMat(&_featureVector);
        _featureVector = NULL;
    }
    if (_weakResponses != NULL) {
        cvReleaseMat(&_weakResponses);
        _weakResponses = NULL;
    }

#if 1
    // erase non-valid samples
    int num_erased = makeSamplesValid(posSamples);
    if (num_erased > 0) SVL_LOG(SVL_LOG_WARNING, "Had to erase " << num_erased << " invalid positive samples");
    num_erased = makeSamplesValid(negSamples);
    if (num_erased > 0) SVL_LOG(SVL_LOG_WARNING, "Had to erase " << num_erased << " invalid negative samples");
    num_erased = makeSamplesValid(posCVsamples);
    if (num_erased > 0) SVL_LOG(SVL_LOG_WARNING, "Had to erase " << num_erased << " invalid positive cv samples");
    num_erased = makeSamplesValid(negCVsamples);
    if (num_erased > 0) SVL_LOG(SVL_LOG_WARNING, "Had to erase " << num_erased << " invalid negative cv samples");
#endif


    SVL_LOG(SVL_LOG_MESSAGE, "Training with " << posSamples.size() << " positive and "
        << negSamples.size() << " negative samples");
    if (posCVsamples.size() > 0) 
      SVL_LOG(SVL_LOG_MESSAGE, "  ...and with " << posCVsamples.size() << " positive and "
	      << negCVsamples.size() << " negative cross-validation samples");
    SVL_LOG(SVL_LOG_MESSAGE, "Feature vector size is " << posSamples[0].size());

#ifdef DEBUG
    ofstream ofs("svlObjectDetector.learnModel.txt");
    for (unsigned i = 0; i < posSamples.size(); i++) {
	for (unsigned j = 0; j < posSamples[i].size(); j++) {
	    ofs << posSamples[i][j] << " ";
	}
	ofs << 1 << endl;
    }
    for (unsigned i = 0; i < negSamples.size(); i++) {
	for (unsigned j = 0; j < negSamples[i].size(); j++) {
	    ofs << negSamples[i][j] << " ";
	}
	ofs << 0 << endl;
    }    
    ofs.close();
#endif

    // assemble design matrix for training
    CvMat *data = cvCreateMat((int)(posSamples.size() + negSamples.size()),
        (int)posSamples[0].size(), CV_32FC1);
    CvMat *labels = cvCreateMat((int)(posSamples.size() + negSamples.size()), 1, CV_32SC1);
    assert((data != NULL) && (labels != NULL));

    for (unsigned i = 0; i < posSamples.size(); i++) {
	SVL_ASSERT_MSG(posSamples[i].size() == (unsigned)data->width, "Failed on "
		       << i << "th sample (out of " << posSamples.size() << "), with "
		       << "data->width: " << data->width << " and posSamples[i].size(): "
		       << posSamples[i].size());

	for (unsigned j = 0; j < (unsigned)data->width; j++) {
	    cvmSet(data, i, j, posSamples[i][j]);
	}
	CV_MAT_ELEM(*labels, int, i, 0) =  1;
    }

    for (unsigned i = 0; i < negSamples.size(); i++) {
      SVL_ASSERT_MSG(negSamples[i].size() == (unsigned)data->width, "Failed on "
		     << i << "th sample (out of " << negSamples.size() << "), with "
		     << "data->width: " << data->width << " and negSamples[i].size(): "
		     << negSamples[i].size());

	for (unsigned j = 0; j < (unsigned)data->width; j++) {
	    cvmSet(data, i + posSamples.size(), j, negSamples[i][j]);
	}
	CV_MAT_ELEM(*labels, int, i + posSamples.size(), 0) = -1;
    }

    CvMat *varType = cvCreateMat(data->width + 1, 1, CV_8UC1);
    for (int i = 0; i < data->width; i++) {
      CV_MAT_ELEM(*varType, unsigned char, i, 0) = CV_VAR_NUMERICAL;
    }
    CV_MAT_ELEM(*varType, unsigned char, data->width, 0) = CV_VAR_CATEGORICAL;


    // train the classifier
    _model = svlSmartPointer<CvBoost>(new CvBoost());
    
    CvBoostParams parameters(CvBoost::GENTLE, getOptionAsInt("boostingRounds"), getOptionAsDouble("trimRate"), \
			     getOptionAsInt("numSplits"), false, NULL);

    SVL_LOG(SVL_LOG_VERBOSE, "Creating boosting model...");
    _model->train(data, CV_ROW_SAMPLE, labels, NULL, NULL, varType, NULL, parameters);
    SVL_LOG(SVL_LOG_VERBOSE, "...done");

    // used if cross-validation is performed
    int num_rounds = getOptionAsInt("boostingRounds");

    float best_score = -1;
    int best_num_rounds = -1;

    while (num_rounds > 0) {    
      CvSlice slice = cvSlice(0, num_rounds);

      SVL_LOG(SVL_LOG_MESSAGE, "\tAfter " << num_rounds << " rounds of training");

      // otherwise the evaluation code will not work, since the number of model
      // parameters might have changed
      if (_weakResponses) {
	cvReleaseMat(&_weakResponses);
	_weakResponses = NULL;
      }

      // evaluate model (on training data) and print the results
      float training_performance = modelPerformance(posSamples, negSamples, "", true, slice);
      
      // if no cross-validation data
      if (posCVsamples.size() == 0) break;
      
      // evaluate model (on cv data)
      float score = modelPerformance(posCVsamples, negCVsamples, "on cv", true, slice);  

      // two things going on here:
      // (1) makes sure that a random spike in cv performance at the beginning of
      //     training doesn't screw it up (has happened before); thus wait either
      //     until 100 rounds have passed, or training performance gets to 1, which
      //     it almost inevitably will (this is hacky, but necessary for now)
      // (2) given the option between two equally-scoring rounds, want to choose
      //     the smaller one, so score >= best_score, not score > best_score
      if ((num_rounds >= 100 || training_performance > 0.99999) &&
	  (score >= best_score)) {
	best_score = score;
	best_num_rounds = num_rounds;
      }
	  
      // with fewer weak learners, each one matters much more
      int num_reeval_rounds = 20;
      if (num_rounds <= 200)
	num_reeval_rounds = 10;
      if (num_rounds <= 100)
	num_reeval_rounds = 5;

      num_rounds -= num_reeval_rounds;
    }


    // release memory
    if (_weakResponses) {
      cvReleaseMat(&_weakResponses);
      _weakResponses = NULL;
    }
    cvReleaseMat(&varType);
    cvReleaseMat(&labels);
    cvReleaseMat(&data);

    //   if cross-validation was being performed, re-train the model with either
    // entire train+cv set, or just the cv set
    //   this is important because the number of boosting rounds chosen
    // is different from the last trained model 
    if (posCVsamples.size() > 0) {
      setOption("boostingRounds", best_num_rounds);
      
      if (retrain) {
	SVL_LOG(SVL_LOG_MESSAGE, "\tRetraining");
	unsigned orig_num_pos = posSamples.size();
	unsigned orig_num_neg = negSamples.size();
	
	Tsamples fullPosSet(orig_num_pos + posCVsamples.size());
	Tsamples fullNegSet(orig_num_neg + negCVsamples.size());
	
	for (unsigned i = 0; i < orig_num_pos; i++)
	  fullPosSet[i] = posSamples[i];
	for (unsigned i = 0; i < posCVsamples.size(); i++)
	  fullPosSet[orig_num_pos + i] = posCVsamples[i];
	
	for (unsigned i = 0; i < orig_num_neg; i++)
	  fullNegSet[i] = negSamples[i];
	for (unsigned i = 0; i < negCVsamples.size(); i++)
	  fullNegSet[orig_num_neg + i] = negCVsamples[i];
	
	learnModel(fullPosSet, fullNegSet);
      }
      else {
	//learnModel(posSamples, negSamples);
	_model->prune(cvSlice(0, best_num_rounds));
      }
    }
}
#endif
